from enum import Enum


class Network(Enum):

    DRIVE = "drive"
    DRIVE_SERVICE = "drive_service"
    WALK = "walk"

    @property
    def filter(self) -> str:
        speeds = {
            Network.DRIVE: (
                f'["highway"]["area"!~"yes"]["access"!~"private"]'
                f'["highway"!~"abandoned|bridleway|bus_guideway|construction|corridor|cycleway|elevator|'
                f"escalator|footway|no|path|pedestrian|planned|platform|proposed|raceway|razed|service|"
                f'steps|track"]'
                f'["motor_vehicle"!~"no"]["motorcar"!~"no"]'
                f'["service"!~"alley|driveway|emergency_access|parking|parking_aisle|private"]'
            ),
            Network.WALK: (
                f'["highway"]["area"!~"yes"]["access"!~"private"]'
                f'["highway"!~"abandoned|bus_guideway|construction|cycleway|motor|no|planned|platform|'
                f'proposed|raceway|razed"]'
                f'["foot"!~"no"]["service"!~"private"]'
                f'["sidewalk"!~"separate"]["sidewalk:both"!~"separate"]'
                f'["sidewalk:left"!~"separate"]["sidewalk:right"!~"separate"]'
            ),
            Network.DRIVE_SERVICE: (
                f'["highway"]["area"!~"yes"]["access"!~"private"]'
                f'["highway"!~"abandoned|bridleway|bus_guideway|construction|corridor|cycleway|elevator|'
                f"escalator|footway|no|path|pedestrian|planned|platform|proposed|raceway|razed|steps|"
                f'track"]'
                f'["motor_vehicle"!~"no"]["motorcar"!~"no"]'
                f'["service"!~"emergency_access|parking|parking_aisle|private"]'
            ),
        }
        return speeds[self]



import math
import warnings
from collections import defaultdict
from itertools import chain, combinations

import numpy as np
import pandas as pd
from pandas import DataFrame
from pyproj import CRS, Transformer
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist
from shapely import LineString, Point
from shapely.ops import substring

from iduedu.modules.overpass_downloaders import fetch_member_tags

PLATFORM_ROLES = ["platform_entry_only", "platform", "platform_exit_only"]
STOPS_ROLES = ["stop", "stop_exit_only", "stop_entry_only"]
THRESHOLD_METERS = 100


def _link_unconnected(disconnected_ways) -> list:
    # Считаем связи между линиями
    connect_points = [point for coords in disconnected_ways for point in (coords[0], coords[-1])]
    distances = cdist(connect_points, connect_points)
    n = distances.shape[0]
    mask = (np.arange(n)[:, None] // 2) == (np.arange(n) // 2)
    distances[mask] = np.inf

    def relative_point(point: int):
        rel_point = point // 2 * 2
        return rel_point if rel_point != point else rel_point + 1

    connected_ways = []
    # pylint: disable=unbalanced-tuple-unpacking
    first_con_1, first_con_2 = np.unravel_index(np.argmin(distances), distances.shape)
    distances[first_con_1, :] = np.inf
    distances[first_con_2, :] = np.inf
    distances[:, first_con_2] = np.inf
    distances[:, first_con_1] = np.inf
    first_con_1_rel, first_con_2_rel = [relative_point(x) for x in (first_con_1, first_con_2)]
    distances[first_con_1_rel, first_con_2_rel] = np.inf
    distances[first_con_2_rel, first_con_1_rel] = np.inf
    distances[:, first_con_1_rel] = np.inf
    distances[:, first_con_2_rel] = np.inf

    line1, line2 = first_con_1 // 2, first_con_2 // 2

    if first_con_1 % 2 == 1:
        connected_ways += disconnected_ways[line1]
    else:
        connected_ways += disconnected_ways[line1][::-1]
    if first_con_2 % 2 == 0:
        connected_ways += disconnected_ways[line2]
    else:
        connected_ways += disconnected_ways[line2][::-1]

    extreme_points = [first_con_1_rel, first_con_2_rel]

    for _ in range(len(disconnected_ways) - 2):
        # pylint: disable=invalid-sequence-index
        position, ind = np.unravel_index(np.argmin(distances[extreme_points]), (2, n))
        next_con = (extreme_points[position], ind)
        rel_point = relative_point(next_con[1])

        line = ind // 2
        if position == 0:
            connected_ways = (
                disconnected_ways[line] + connected_ways
                if ind % 2 == 1
                else (disconnected_ways[line][::-1] + connected_ways)
            )
            extreme_points = [rel_point, extreme_points[1]]
        else:
            connected_ways = (
                connected_ways + disconnected_ways[line]
                if ind % 2 == 0
                else (connected_ways + disconnected_ways[line][::-1])
            )
            extreme_points = [extreme_points[0], rel_point]

        # Чтоб нельзя было зациклиться
        distances[:, rel_point] = np.inf
        distances[next_con[0], :] = np.inf
        distances[next_con[1], :] = np.inf
        distances[:, next_con[0]] = np.inf
        distances[:, next_con[1]] = np.inf
        rel_point_2 = relative_point(next_con[0])
        distances[rel_point, rel_point_2] = np.inf
        distances[rel_point_2, rel_point] = np.inf

    return connected_ways


def parse_overpass_route_response(loc: dict, crs: CRS, needed_tags: list[str]) -> pd.Series:
    transformer = Transformer.from_crs("EPSG:4326", crs, always_xy=True)

    # TODO надо писать версию с merge_line и отдельными запросами по максимальной скорости
    def transform_geometry(loc):
        if isinstance(loc["geometry"], float):
            return transformer.transform(loc["lon"], loc["lat"])
        p = LineString([transformer.transform(coords["lon"], coords["lat"]) for coords in loc["geometry"]]).centroid
        return p.x, p.y

    def process_roles(route, roles):
        filtered = route[route["role"].isin(roles)]
        return filtered.apply(transform_geometry, axis=1).tolist(), filtered["ref"].tolist()

    def extract_needed(loc_obj: dict, keys: list[str]) -> dict | None:
        out = {}
        tags = loc_obj.get("tags", {}) if isinstance(loc_obj.get("tags"), dict) else {}
        for k in keys:
            value_found = None
            if "." in k:
                # dotted path
                parts = k.split(".")
                cur = loc_obj
                ok = True
                for part in parts:
                    if isinstance(cur, dict) and part in cur:
                        cur = cur[part]
                    else:
                        ok = False
                        break
                if ok:
                    value_found = cur
            else:
                if k in tags:
                    value_found = tags[k]
                elif k in loc_obj:
                    value_found = loc_obj[k]
            if value_found is not None:
                out[k] = value_found
        return out or None

    tags = loc.get("tags", {}) if isinstance(loc.get("tags"), dict) else {}

    transport_name = None
    if "ref" in tags:
        transport_name = tags["ref"]
    elif "name" in tags:
        transport_name = tags["name"]

    members = loc.get("members", [])
    route = pd.DataFrame(members) if isinstance(members, list) else pd.DataFrame()

    for col in ("geometry", "lat", "lon", "role", "type"):
        if col not in route.columns:
            route[col] = np.nan

    route = route.dropna(subset=["lat", "lon", "geometry"], how="all")

    platforms, platforms_refs = process_roles(route, PLATFORM_ROLES)
    stops, stops_refs = process_roles(route, STOPS_ROLES)

    ways_df = route[(route["type"] == "way") & (route["role"].fillna("") == "")]

    if not ways_df.empty:
        ways = (
            ways_df["geometry"]
            .reset_index(drop=True)
            .apply(lambda g: [transformer.transform(pt["lon"], pt["lat"]) for pt in g])
            .tolist()
        )
        connected_ways = [[]]
        cur_way = 0
        for coords in ways:
            if not coords:
                continue
            # Соединяем маршруты, если всё ок, идут без пропусков
            if not connected_ways[cur_way]:
                connected_ways[cur_way] += coords
                continue
            if coords[0] == coords[-1]:
                # Круговое движение зацикленное зачастую в осм, можно отработать, но сходу не придумал
                continue
            if connected_ways[cur_way][-1] == coords[0]:
                connected_ways[cur_way] += coords[1:]

            elif connected_ways[cur_way][-1] == coords[-1]:
                connected_ways[cur_way] += coords[::-1][1:]

            elif connected_ways[cur_way][0] == coords[0]:
                connected_ways[cur_way] = coords[1:][::-1] + connected_ways[cur_way]

            elif connected_ways[cur_way][0] == coords[-1]:
                connected_ways[cur_way] = coords + connected_ways[cur_way][1:]

            # Случай если нету соединяющей точки между линиями маршрута
            else:
                connected_ways += [coords]
                cur_way += 1
        # Соединяем линии по ближайшим точкам этих линий

        if len(connected_ways) > 1:
            # удаляем все круговые движения
            to_del = [i for i, data in enumerate(connected_ways) if data[0] == data[-1]]
            # Если кол-во удалений == кол-во путей, надо оставить хотя бы самый большой
            if len(to_del) == len(connected_ways):
                # Найдём индекс самого длинного пути среди замкнутых
                longest_index = max(to_del, key=lambda i: len(connected_ways[i]))
                # Удалим все, кроме самого длинного
                to_del.remove(longest_index)
            connected_ways = [i for j, i in enumerate(connected_ways) if j not in to_del]
        if len(connected_ways) > 1:
            path = _link_unconnected(connected_ways)
        else:
            if not connected_ways or not connected_ways[0]:
                raise Exception("No connected ways")
            path = connected_ways[0]

    else:
        path = None

    extra = extract_needed(loc, needed_tags)

    return pd.Series(
        {
            "path": path,
            "platforms": platforms,
            "platforms_refs": platforms_refs,
            "stops": stops,
            "stops_refs": stops_refs,
            "route": transport_name,
            "extra_data": extra,
        }
    )


def geometry_to_graph_edge_node_df(loc: pd.Series, transport_type, loc_id) -> DataFrame | None:

    name = loc.route

    platforms = list(loc.platforms or [])  # same size ->
    platforms_refs = list(loc.platforms_refs or [])

    stops = list(loc.stops or [])
    stops_refs = list(loc.stops_refs or [])

    path = loc.path
    extra_data = loc.extra_data

    if not path:
        warnings.warn("В одном из маршрутов нет пути, пока не работаю с этим", UserWarning)
        return None

    def _side_left_or_right(point):
        # 1 if left 0 if right для определения с какой стороны от линии точки
        dist = path.project(point)

        d1 = dist - 1 if dist - 1 > 0 else 0
        d2 = dist + 1 if dist + 1 < path.length else path.length
        line = substring(path, d1, d2)
        x1, y1 = line.coords[0]
        x2, y2 = line.coords[-1]
        x, y = point.coords[0]

        cross_product = (x2 - x1) * (y - y1) - (y2 - y1) * (x - x1)
        if cross_product > 0:
            return 1
        return 0

    def _offset_point(point, direction, distance=7) -> tuple[float, float]:
        # для размещения платформы по одну сторону от пути на расстоянии
        dist = path.project(point)
        d1 = dist - 1 if dist - 1 > 0 else 0
        d2 = dist + 1.1 if dist + 1.1 < path.length else path.length
        nearest_pt_on_line = path.interpolate(dist)
        line = substring(path, d1, d2)

        x1, y1 = line.coords[0]
        x2, y2 = line.coords[-1]

        dx, dy = x2 - x1, y2 - y1
        length = math.sqrt(dx**2 + dy**2)
        dx, dy = dx / length, dy / length

        if direction == 0:  # Вправо
            nx, ny = dy, -dx
        else:  # Влево
            nx, ny = -dy, dx

        # Смещенная точка
        offset_x = nearest_pt_on_line.x + nx * distance
        offset_y = nearest_pt_on_line.y + ny * distance
        return offset_x, offset_y

    path = LineString(path)

    items = []  # [(platform_coords, platform_ref, stop_coords, stop_ref), на выходе не может быть пары без платформы

    platform_len, stops_len = len(platforms), len(stops)
    matched_p, matched_s = set(), set()
    pairs = {}  # p_idx -> s_idx

    if platform_len and stops_len:

        stop_tree = cKDTree(stops)
        k = min(2, stops_len)
        dists, idxs = stop_tree.query(platforms, k=k)

        dists = np.asarray(dists)
        idxs = np.asarray(idxs)
        if dists.ndim == 1:  # это случается, когда k == 1
            dists = dists[:, None]
            idxs = idxs[:, None]

        candidates = []
        for p_i in range(platform_len):
            for rank in range(dists.shape[1]):
                dist = float(dists[p_i, rank])
                if not np.isfinite(dist) or dist > THRESHOLD_METERS:
                    continue
                s_i = int(idxs[p_i, rank])
                candidates.append((dist, p_i, s_i))

        candidates.sort(key=lambda t: t[0])
        for dist, p_i, s_i in candidates:
            if p_i not in matched_p and s_i not in matched_s:
                pairs[p_i] = s_i
                matched_p.add(p_i)
                matched_s.add(s_i)

    for p_i, s_i in pairs.items():
        items.append(
            {
                "p": platforms[p_i],
                "pref": platforms_refs[p_i],
                "s": stops[s_i],
                "sref": stops_refs[s_i],
            }
        )

    if stops_len:
        base_dir = _side_left_or_right(Point(platforms[platform_len // 2])) if platform_len else 1
        for s_i in range(stops_len):
            if s_i in matched_s:
                continue
            s_pt = Point(stops[s_i])
            items.append(
                {
                    "p": _offset_point(s_pt, base_dir),
                    "pref": f"from_{stops_refs[s_i]}",  # new platform
                    "s": s_pt.xy,
                    "sref": stops_refs[s_i],
                }
            )

    for p_i in range(platform_len):
        if p_i in matched_p:
            continue
        items.append(
            {
                "p": platforms[p_i],
                "pref": platforms_refs[p_i],
                "s": None,
                "sref": f"from_{platforms_refs[p_i]}",
            }
        )

    if not items:
        items.append({"p": _offset_point(path.interpolate(0), 1, 7), "pref": None, "s": None, "sref": None})
        items.append({"p": _offset_point(path.interpolate(path.length), 1, 7), "pref": None, "s": None, "sref": None})

    items.sort(key=lambda d: path.project(Point(d["p"])))

    graph_data: list[dict] = []
    node_id = 0
    last_dist = None
    last_projected_stop_id = None

    def add_node(x, y, node_type=None, ref_id=None):
        x = round(x, 5)
        y = round(y, 5)
        graph_data.append(
            {
                "node_id": (loc_id, node_id),
                "point": (x, y),
                "route": name,
                "type": (node_type if node_type else transport_type),
                "ref_id": ref_id,
            }
        )

    def add_edge(u, v, geometry=None, transport=None):

        payload = {
            "u": (loc_id, u),
            "v": (loc_id, v),
            "route": name,
            "type": (transport_type if transport else "boarding"),
            "extra_data": extra_data,
        }
        if geometry is not None:
            payload["geometry"] = LineString([(round(x, 5), round(y, 5)) for x, y in geometry.coords])
        graph_data.append(payload)

    for item in items:
        p_xy = item["p"]
        p_ref = item.get("pref")

        s_xy = item["s"]
        if s_xy is None:
            s_xy = p_xy
        s_ref = item.get("sref")

        platform = Point(p_xy)
        stop = Point(s_xy)

        stop_dist = path.project(stop)

        projected_stop = path.interpolate(stop_dist)
        add_node(projected_stop.x, projected_stop.y, ref_id=s_ref)

        if last_dist is not None:
            seg = substring(path, last_dist, stop_dist)
            if isinstance(seg, Point):
                seg = LineString((seg, seg))
            add_edge(last_projected_stop_id, node_id, geometry=seg, transport=True)

        last_projected_stop_id = node_id
        last_dist = stop_dist
        node_id += 1

        add_node(platform.x, platform.y, node_type="platform", ref_id=p_ref)
        boarding_geom = LineString(
            [(round(projected_stop.x, 5), round(projected_stop.y, 5)), (round(platform.x, 5), round(platform.y, 5))]
        )
        add_edge(node_id - 1, node_id, geometry=boarding_geom)
        add_edge(node_id, node_id - 1, geometry=boarding_geom)
        node_id += 1

    to_return = pd.DataFrame(graph_data)
    return to_return


def parse_overpass_to_edgenode(loc, crs, needed_tags) -> pd.DataFrame | None:
    loc_id = loc.name
    transport_type = loc.transport_type
    parsed_geometry = parse_overpass_route_response(loc, crs, needed_tags)
    edgenode = geometry_to_graph_edge_node_df(parsed_geometry, transport_type, loc_id)
    return edgenode


def infer_role_from_tags(tags: dict) -> str:
    _TRUE = {"yes", "true", "1"}

    def _is_true(v):
        return str(v).lower() in _TRUE

    if not tags:
        return ""

    pt = (tags.get("public_transport", "")).lower()
    rail = (tags.get("railway", "")).lower()
    stat = (tags.get("station", "")).lower()
    entr = (tags.get("entrance", "")).lower()
    entry = (tags.get("entry", "")).lower()
    exit_ = (tags.get("exit", "")).lower()

    # station
    if pt == "station" or rail == "station" or stat == "subway":
        return "station"

    # platform
    if pt == "platform" or rail == "platform":
        return "platform"

    # stop / stop_position
    if pt in {"stop_position", "stop"} or rail in {"stop", "halt"}:
        return "stop"

    # entrances
    if rail == "subway_entrance" or entr in {"yes", "main", "service", "entrance", "entry", "exit"}:
        if entr == "entry" or (_is_true(entry) and not _is_true(exit_)):
            return "entry_only"
        if entr == "exit" or (_is_true(exit_) and not _is_true(entry)):
            return "exit_only"
        return "entrance"
    return ""


def patch_members_roles_inplace(stop_areas_df):

    missing = []
    for _, r in stop_areas_df.iterrows():
        for m in r.get("members") or []:
            role = (m.get("role") or "").strip()
            if role == "":
                missing.append({"type": m["type"], "ref": int(m["ref"])})

    if not missing:
        return

    tags_map = fetch_member_tags(missing)

    for _, r in stop_areas_df.iterrows():
        for m in r.get("members") or []:
            if (m.get("role") or "").strip():
                continue
            key = (m["type"], int(m["ref"]))
            tags = tags_map.get(key, {})
            role = infer_role_from_tags(tags)
            if role:
                m["role"] = role
            if "lon" not in m and "__center__" in tags:
                cx, cy = tags["__center__"]
                m["lon"], m["lat"] = cx, cy


def parse_overpass_subway_data(
    stop_areas, stop_areas_group, stations_data, to_crs
) -> tuple[pd.DataFrame, pd.DataFrame]:
    graph_nodes = []
    graph_edges = []

    transformer = Transformer.from_crs("EPSG:4326", to_crs, always_xy=True)

    station_parent_ref = {}
    osm_roles_to_iduedu = {
        "station": "subway_station",
        "stop": "subway",
        "platform": "subway_platform",
        "entrance": "subway_entry_exit",
        "subway_entrance": "subway_entry_exit",
        "entrance_yes": "subway_entry_exit",
        "entrance_main": "subway_entry_exit",
        "entry_only": "subway_entry",
        "entrance_entry_only": "subway_entry",
        "entry": "subway_entry",
        "exit_only": "subway_exit",
        "entrance_exit_only": "subway_exit",
        "exit": "subway_exit",
    }

    def _collect_by_roles(members, parent_id):
        roles = defaultdict(list)
        for m in members or []:
            role = (m.get("role") or "").lower()
            roles[role].append(m)

        stations = roles.get("station", [])
        for s in stations:
            station_parent_ref[parent_id] = s["ref"]
        stops = roles.get("stop", [])
        platforms = roles.get("platform", [])

        entrances = (
            roles.get("entrance", [])
            + roles.get("subway_entrance", [])
            + roles.get("entrance_yes", [])
            + roles.get("entrance_main", [])
        )
        entry_only = roles.get("entry_only", []) + roles.get("entrance_entry_only", []) + roles.get("entry", [])
        exit_only = roles.get("exit_only", []) + roles.get("entrance_exit_only", []) + roles.get("exit", [])

        return stations, stops, platforms, entrances, entry_only, exit_only

    def add_node(ref_id, x, y, node_type):

        if x is None or y is None:
            point = None
        else:
            point = transformer.transform(x, y)

        graph_nodes.append(
            {
                "ref_id": int(ref_id),
                "point": point,
                "type": osm_roles_to_iduedu.get(node_type),
            }
        )

    def add_edge(u, v, edge_type):
        graph_edges.append(
            {
                "u_ref": int(u),
                "v_ref": int(v),
                "type": edge_type,
            }
        )

    patch_members_roles_inplace(stop_areas)

    for _, stop_area in stop_areas.iterrows():

        members = stop_area["members"]
        all_nodes = _collect_by_roles(members, stop_area["id"])
        start_idx = len(graph_nodes)

        for node in chain.from_iterable(all_nodes):
            if "geometry" in node:
                new_lon, new_lat = LineString((xy["lon"], xy["lat"]) for xy in node["geometry"]).centroid.xy
                node["lon"], node["lat"] = new_lon[0], new_lat[0]
            if "lon" not in node:
                node["lon"], node["lat"] = None, None

            add_node(node["ref"], node["lon"], node["lat"], node["role"])

        ref2idx = {graph_nodes[i]["ref_id"]: i for i in range(start_idx, len(graph_nodes))}

        stations, stops, platforms, entrances, entry_only, exit_only = all_nodes

        for stop in stops:
            for platform in platforms:
                add_edge(stop["ref"], platform["ref"], "boarding")
                add_edge(platform["ref"], stop["ref"], "boarding")

        has_station = len(stations) > 0

        if has_station:
            for platform in platforms:
                for station in stations:
                    add_edge(station["ref"], platform["ref"], "subway_station")
                    add_edge(platform["ref"], station["ref"], "subway_station")

        targets = [s["ref"] for s in stations] if has_station else [p["ref"] for p in platforms]

        has_entrance = (len(entrances) + len(entry_only)) > 0
        has_exit = (len(entrances) + len(exit_only)) > 0

        for entrance in entrances:
            for t in targets:
                add_edge(entrance["ref"], t, "subway_entrance")
                add_edge(t, entrance["ref"], "subway_exit")

        for entrance in entry_only:
            for t in targets:
                add_edge(entrance["ref"], t, "subway_entrance")

        for entrance in exit_only:
            for t in targets:
                add_edge(t, entrance["ref"], "subway_exit")

        if not (has_entrance and has_exit):
            for t in targets:
                idx = ref2idx.get(int(t))
                if idx is not None:
                    graph_nodes[idx]["type"] = "platform"

    for _, stop_area_group in stop_areas_group.iterrows():
        members = stop_area_group["members"]

        connect_refs = [station_parent_ref[mem["ref"]] for mem in members if mem["ref"] in station_parent_ref]

        pairs = list(combinations(connect_refs, 2))
        for pair in pairs:
            if pair[0] != pair[1]:
                add_edge(pair[0], pair[1], "subway_transfer")
                add_edge(pair[1], pair[0], "subway_transfer")

    edges_gdf = pd.DataFrame(graph_edges)
    nodes_gdf = pd.DataFrame(graph_nodes)
    nodes_gdf["extra_data"] = {}

    for _, station_data in stations_data.iterrows():
        tags = station_data["tags"]
        ref_id = station_data["id"]
        station_ind = nodes_gdf[nodes_gdf["ref_id"] == int(ref_id)].index
        extra_data = {k: v for k, v in tags.items() if k in ["name", "depth"]}
        extra_data["depth"] = extra_data.get("depth", 0)
        nodes_gdf.loc[station_ind, "extra_data"] = [extra_data for _ in station_ind]

    edges_gdf = edges_gdf.drop_duplicates(subset=["u_ref", "v_ref"])
    edges_gdf["length_meter"] = 0.0
    edges_gdf["time_min"] = 0.0
    edges_gdf["geometry"] = None

    def _depth_from_extra(d):
        if isinstance(d, dict):
            v = d.get("depth", 0)
            try:
                return float(v)
            except (TypeError, ValueError):
                return 0.0
        return 0.0

    nodes_info = nodes_gdf[["ref_id", "point", "type", "extra_data"]].copy()
    nodes_info["depth_m"] = nodes_info["extra_data"].apply(_depth_from_extra)

    nodes_lut = nodes_info.set_index("ref_id")[["point", "type", "depth_m"]]

    edges_nodes = (
        edges_gdf.join(nodes_lut.add_prefix("u_"), on="u_ref")
        .join(nodes_lut.add_prefix("v_"), on="v_ref")
        .drop_duplicates(subset=["u_ref", "v_ref"])
    )

    def _horiz_len(pu, pv):
        if pu is None or pv is None:
            return np.nan
        try:
            return LineString([pu, pv]).length
        except Exception:
            return np.nan

    edges_nodes["horiz_m"] = edges_nodes.apply(lambda r: _horiz_len(r["u_point"], r["v_point"]), axis=1)

    edges_nodes["station_depth_m"] = np.where(
        edges_nodes["u_type"].eq("subway_station"),
        edges_nodes["u_depth_m"].fillna(0.0),
        np.where(edges_nodes["v_type"].eq("subway_station"), edges_nodes["v_depth_m"].fillna(0.0), 0.0),
    ).astype(float)

    ESCALATOR_SPEED_MPMIN = 0.75 * 60
    WALK_SPEED_MPMIN = 5 * 1000 / 60
    TRANSFER_PATH_FACTOR = 1.5

    mask_escal = edges_nodes["type"].isin(["subway_entrance", "subway_exit"]) & edges_nodes["horiz_m"].notna()
    edges_nodes.loc[mask_escal, "length_meter"] = np.hypot(
        edges_nodes.loc[mask_escal, "horiz_m"], edges_nodes.loc[mask_escal, "station_depth_m"]
    )
    edges_nodes.loc[mask_escal, "time_min"] = edges_nodes.loc[mask_escal, "length_meter"] / ESCALATOR_SPEED_MPMIN

    def _straight_geom(pu, pv):
        if pu is None or pv is None:
            return None
        try:
            return LineString([pu, pv])
        except Exception:
            return None

    edges_nodes.loc[mask_escal, "geometry"] = edges_nodes.loc[mask_escal].apply(
        lambda r: _straight_geom(r["u_point"], r["v_point"]), axis=1
    )

    mask_transfer = edges_nodes["type"].eq("subway_transfer") & edges_nodes["horiz_m"].notna()
    edges_nodes.loc[mask_transfer, "length_meter"] = TRANSFER_PATH_FACTOR * edges_nodes.loc[
        mask_transfer, "horiz_m"
    ].round(1)
    edges_nodes.loc[mask_transfer, "time_min"] = (
        edges_nodes.loc[mask_transfer, "length_meter"] / WALK_SPEED_MPMIN
    ).round(1)
    edges_nodes.loc[mask_transfer, "geometry"] = edges_nodes.loc[mask_transfer].apply(
        lambda r: _straight_geom(r["u_point"], r["v_point"]), axis=1
    )

    edges_gdf[["length_meter", "time_min", "geometry"]] = edges_nodes[["length_meter", "time_min", "geometry"]]
    return edges_gdf, nodes_gdf



import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from loguru import logger
from pyproj import CRS
from pyproj.aoi import AreaOfInterest

# pylint: disable=no-name-in-module
from pyproj.database import query_utm_crs_info
from shapely import LineString, MultiLineString, Point, Polygon, from_wkt, line_merge, node
from shapely.geometry.base import BaseGeometry


def clip_nx_graph(graph: nx.Graph, polygon: Polygon) -> nx.Graph:
    """
    Clip a NetworkX graph by a polygon and return the induced subgraph.

    Builds a GeoDataFrame of node points using `graph.graph["crs"]`, clips it by the given
    polygon, then returns the node-induced subgraph (keeping original node/edge attributes).

    Parameters:
        graph (nx.Graph): Graph with node coords stored as `x`, `y` and a CRS in `graph.graph["crs"]`.
        polygon (Polygon): Clipping polygon in the same CRS as the graph.

    Returns:
        (nx.Graph): Subgraph containing only nodes whose points fall inside the polygon.

    Notes:
        Edges are preserved only if both endpoints remain in the subgraph.
    """
    crs = graph.graph["crs"]
    points = gpd.GeoDataFrame(
        data=[{"id": p_id, "geometry": Point(data["x"], data["y"])} for p_id, data in graph.nodes(data=True)], crs=crs
    ).clip(polygon, True)
    clipped = graph.subgraph(points["id"].tolist())
    return clipped


def _fmt_top_sizes(sizes, top_k: int = 5) -> str:
    ss = sorted(sizes, reverse=True)
    if len(ss) <= top_k:
        return "[" + ", ".join(map(str, ss)) + "]"
    return "[" + ", ".join(map(str, ss[:top_k])) + ", …]"


def keep_largest_strongly_connected_component(graph: nx.DiGraph, *, top_k_wcc_sizes: int = 5) -> nx.DiGraph:
    """
    Keep only the largest strongly connected component of a directed graph.

    Logs the sizes of weakly connected components (WCC) for visibility, then removes all
    nodes outside the largest strongly connected component (SCC) and returns the pruned copy.

    Parameters:
        graph (nx.DiGraph): Directed graph to prune (a copy is made).
        top_k_wcc_sizes (int): How many largest WCC sizes to show in the warning.

    Returns:
        (nx.DiGraph): Graph restricted to the largest SCC.

    Notes:
        - Uses `nx.weakly_connected_components` for a quick disconnectedness summary.
        - Nodes from all SCCs except the largest are removed.
    """
    graph = graph.copy()

    weakly_connected_components = list(nx.weakly_connected_components(graph))
    if len(weakly_connected_components) > 1:
        sizes = [len(c) for c in weakly_connected_components]
        logger.warning(
            f"Graph contains {len(weakly_connected_components)} weakly connected components. "
            f"This means the graph has disconnected groups if edge directions are ignored. "
            f"Component sizes:: {_fmt_top_sizes(sizes, top_k=top_k_wcc_sizes)}"
        )

    all_scc = sorted(nx.strongly_connected_components(graph), key=len)
    nodes_to_del = set().union(*all_scc[:-1])

    if nodes_to_del:
        logger.warning(
            f"Removing {len(nodes_to_del)} nodes from {len(all_scc) - 1} smaller strongly connected components. "
            f"These are subgraphs where nodes are internally reachable but isolated from the rest. "
            f"Retaining only the largest strongly connected component ({len(all_scc[-1])} nodes)."
        )
        graph.remove_nodes_from(nodes_to_del)

    return graph


def estimate_crs_for_bounds(minx, miny, maxx, maxy) -> CRS:
    """
    Estimate a local UTM CRS for the given lon/lat bounds.

    Uses the bounds center to query a suitable WGS84 UTM CRS via `pyproj.query_utm_crs_info`
    and returns it as a `pyproj.CRS`.

    Parameters:
        minx (float): Western longitude.
        miny (float): Southern latitude.
        maxx (float): Eastern longitude.
        maxy (float): Northern latitude.

    Returns:
        (pyproj.CRS): UTM CRS suited for metric distance calculations near the bounds center.
    """
    x_center = np.mean([minx, maxx])
    y_center = np.mean([miny, maxy])
    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(
            west_lon_degree=x_center,
            south_lat_degree=y_center,
            east_lon_degree=x_center,
            north_lat_degree=y_center,
        ),
    )
    crs = CRS.from_epsg(utm_crs_list[0].code)
    logger.debug(f"Estimated CRS for territory {crs}")
    return crs


def _edges_to_gdf(graph: nx.Graph, crs) -> gpd.GeoDataFrame:
    """
    Converts nx graph to gpd.GeoDataFrame as edges.
    """
    graph_df = pd.DataFrame(list(graph.edges(data=True)), columns=["u", "v", "data"])
    edge_data_expanded = pd.json_normalize(graph_df["data"])
    graph_df = pd.concat([graph_df.drop(columns=["data"]), edge_data_expanded], axis=1)
    graph_df = gpd.GeoDataFrame(graph_df, geometry="geometry", crs=crs).set_index(["u", "v"])
    graph_df["geometry"] = graph_df["geometry"].fillna(LineString())
    return graph_df


def _nodes_to_gdf(graph: nx.Graph, crs: int) -> gpd.GeoDataFrame:
    """
    Converts nx graph to gpd.GeoDataFrame as nodes.
    """

    ind, data = zip(*graph.nodes(data=True))
    node_geoms = (Point(d["x"], d["y"]) for d in data)
    gdf_nodes = gpd.GeoDataFrame(data, index=ind, crs=crs, geometry=list(node_geoms))

    return gdf_nodes


def _restore_edges_geom(nodes_gdf, edges_gdf) -> gpd.GeoDataFrame:
    edges_wout_geom = edges_gdf[edges_gdf["geometry"].is_empty].reset_index()
    edges_wout_geom["geometry"] = [
        LineString((s, e))
        for s, e in zip(
            nodes_gdf.loc[edges_wout_geom["u"], "geometry"], nodes_gdf.loc[edges_wout_geom["v"], "geometry"]
        )
    ]
    edges_wout_geom.set_index(["u", "v"], inplace=True)
    edges_gdf.update(edges_wout_geom)
    return edges_gdf


def graph_to_gdf(
    graph: nx.MultiDiGraph, edges: bool = True, nodes: bool = True, restore_edge_geom=False
) -> gpd.GeoDataFrame | None:
    """
    Convert a NetworkX graph to GeoDataFrames (edges and/or nodes).

    Reads CRS from `graph.graph["crs"]`. Depending on flags, returns only nodes, only edges,
    or a concatenation of both. Optionally reconstructs missing edge geometries from node points.

    Parameters:
        graph (nx.MultiDiGraph): Graph with node coords (`x`, `y`) and optional edge `geometry`.
        edges (bool): If True, include edges.
        nodes (bool): If True, include nodes.
        restore_edge_geom (bool): If True, fill empty edge geometries from node coordinates.

    Returns:
        (gpd.GeoDataFrame | None): Nodes and/or edges as GeoDataFrame(s).
            If both `edges` and `nodes` are False, returns None.

    Raises:
        ValueError: If `graph.graph["crs"]` is missing.

    Notes:
        - Edge GeoDataFrame uses MultiIndex `(u, v)`.
        - When both are requested, the result is a vertical concat of nodes then edges.
    """
    try:
        crs = graph.graph["crs"]
    except KeyError as exc:
        raise ValueError("Graph does not have crs attribute and no crs was provided") from exc
    if not edges and not nodes:
        logger.debug("Neither edges or nodes were selected, graph_to_gdf returning None")
        return None
    if nodes and not edges:
        nodes_gdf = _nodes_to_gdf(graph, crs)
        return nodes_gdf
    if not nodes and edges:
        edges_gdf = _edges_to_gdf(graph, crs)
        if restore_edge_geom:
            nodes_gdf = _nodes_to_gdf(graph, crs)
            edges_gdf = _restore_edges_geom(nodes_gdf, edges_gdf)
        return edges_gdf

    nodes_gdf = _nodes_to_gdf(graph, crs)
    edges_gdf = _edges_to_gdf(graph, crs)
    if restore_edge_geom:
        edges_gdf = _restore_edges_geom(nodes_gdf, edges_gdf)
    full_gdf = pd.concat([nodes_gdf, edges_gdf])
    return full_gdf


def gdf_to_graph(
    gdf: gpd.GeoDataFrame, project_gdf_attr=True, reproject_to_utm_crs=True, speed=5, check_intersections=True
) -> nx.DiGraph:
    """
    Convert a GeoDataFrame of LineStrings into a directed graph (nx.DiGraph).

    Explodes multilines, optionally enforces topological intersections, merges collinear segments,
    transfers selected attributes back to merged lines via centroid-buffer spatial join,
    and constructs a directed graph whose edges correspond to line segments. Lengths are computed
    in meters in a local metric CRS; travel time uses a provided speed.

    Parameters:
        gdf (gpd.GeoDataFrame): Input with LineString geometries (other types are filtered out).
        project_gdf_attr (bool): If True, projects original attributes to merged lines via nearest overlay.
        reproject_to_utm_crs (bool): If True, lengths computed in UTM and optionally reprojected back.
        speed (float): Speed in km/h used to compute `time_min` for each edge.
        check_intersections (bool): If True, uses `shapely.node` before `line_merge` to enforce proper splits.

    Returns:
        (nx.DiGraph): Directed graph with:
            - node attributes: `x`, `y`;
            - edge attributes: `geometry`, `length_meter`, `time_min`, plus projected attributes;

            Graph attribute `graph["crs"]` is set to the (possibly reprojected) CRS.

    Raises:
        ValueError: If the input contains no valid LineStrings.

    Notes:
        - Attribute projection aggregates multi-matches via a uniqueness reducer (`unique_list`).
        - `speed` is internally converted to meters/minute.
    """

    def unique_list(agg_vals):
        agg_vals = list(set(agg_vals.dropna()))
        if len(agg_vals) == 1:
            return agg_vals[0]
        return agg_vals

    original_crs = gdf.crs
    gdf = gdf.to_crs(gdf.estimate_utm_crs())

    gdf = gdf.explode(ignore_index=True)
    gdf = gdf[gdf.geom_type == "LineString"]

    if len(gdf) == 0:
        raise ValueError("Provided GeoDataFrame contains no valid LineStrings")

    if check_intersections:
        lines = line_merge(node(MultiLineString(gdf.geometry.to_list())))
    else:
        lines = line_merge(MultiLineString(gdf.geometry.to_list()))

    if isinstance(lines, LineString):
        lines = MultiLineString([lines])

    lines = gpd.GeoDataFrame(geometry=list(lines.geoms), crs=gdf.crs)

    if len(gdf.columns) > 1 and project_gdf_attr:
        lines_centroids = lines.copy()
        lines_centroids.geometry = lines_centroids.apply(
            lambda row: row.geometry.line_interpolate_point(row.geometry.length / 2), axis=1
        ).buffer(0.05, resolution=2)
        lines_with_attrs = gpd.sjoin(lines_centroids, gdf, how="left", predicate="intersects")
        aggregated_attrs = (
            lines_with_attrs.drop(columns=["geometry", "index_right"])  # убрать геометрию буфера
            .groupby(lines_with_attrs.index)
            .agg(unique_list)
        )
        lines = pd.concat([lines, aggregated_attrs], axis=1)

    lines["length_meter"] = np.round(lines.length, 2)
    if not reproject_to_utm_crs:
        lines = lines.to_crs(original_crs)

    coords = lines.geometry.get_coordinates()
    coords_grouped_by_index = coords.reset_index(names="old_index").groupby("old_index")
    start_coords = coords_grouped_by_index.head(1).apply(lambda a: (a.x, a.y), axis=1).rename("start")
    end_coords = coords_grouped_by_index.tail(1).apply(lambda a: (a.x, a.y), axis=1).rename("end")
    coords = pd.concat([start_coords.reset_index(), end_coords.reset_index()], axis=1)[["start", "end"]]
    lines = pd.concat([lines, coords], axis=1)
    unique_coords = pd.concat([coords["start"], coords["end"]], ignore_index=True).unique()
    coord_to_index = {coord: idx for idx, coord in enumerate(unique_coords)}

    lines["u"] = lines["start"].map(coord_to_index)
    lines["v"] = lines["end"].map(coord_to_index)

    speed = speed * 1000 / 60
    lines["time_min"] = np.round(lines["length_meter"] / speed, 2)

    graph = nx.Graph()
    for coords, node_id in coord_to_index.items():
        x, y = coords
        graph.add_node(node_id, x=float(x), y=float(y))

    columns_to_attr = lines.columns.difference(["start", "end", "u", "v"])
    for _, row in lines.iterrows():
        edge_attrs = {}
        for col in columns_to_attr:
            edge_attrs[col] = row[col]
        graph.add_edge(row.u, row.v, **edge_attrs)

    graph.graph["crs"] = lines.crs
    graph.graph["speed m/min"] = speed
    return nx.DiGraph(graph)


def write_gml(graph: nx.Graph, gml_path: str) -> nx.Graph:
    """
    Write a NetworkX graph to GML, coercing node coordinates to plain floats.

    Ensures node attributes `x` and `y` are Python `float`, then writes the graph using
    `stringizer=str` so any non-primitive attribute values are serialized as strings.

    Parameters:
        graph (nx.Graph): Input graph. Not mutated — a sanitized copy is written.
        gml_path (str): Output GML file path.

    Returns:
        (nx.Graph): The sanitized copy of the graph that was written to disk.
    """
    graph = graph.copy()

    # Nodes: x/y to float; node geometry to WKT if present
    for n, data in graph.nodes(data=True):
        if "x" in data:
            try:
                data["x"] = float(data["x"])
            except Exception:
                raise ValueError(f"Node {n} has non-numeric x={data['x']!r}")
        if "y" in data:
            try:
                data["y"] = float(data["y"])
            except Exception:
                raise ValueError(f"Node {n} has non-numeric y={data['y']!r}")

    nx.write_gml(graph, gml_path, stringizer=lambda v: str(v))
    return graph


def read_gml(gml_path: str, **nx_kwargs) -> nx.Graph:
    """
    Read a GML file into a NetworkX graph and cast edge `geometry` from WKT strings to shapely.

    Loads the graph via `networkx.read_gml` and, when an edge attribute `geometry` is a string,
    attempts to parse it with `shapely.wkt.from_wkt`. Non-parsable strings are left unchanged.

    Parameters:
        gml_path (str): Path to the GML file.
        **nx_kwargs: Additional keyword arguments forwarded to `networkx.read_gml`.

    Returns:
        (nx.Graph): The graph with edge `geometry` parsed to shapely objects where possible.
    """
    graph = nx.read_gml(gml_path, **nx_kwargs)

    if graph.is_multigraph():
        for u, v, k, data in graph.edges(keys=True, data=True):
            if "geometry" in data and isinstance(data["geometry"], str):
                try:
                    data["geometry"] = from_wkt(data["geometry"])
                except Exception:
                    pass
    else:
        for u, v, data in graph.edges(data=True):
            if "geometry" in data and isinstance(data["geometry"], str):
                try:
                    data["geometry"] = from_wkt(data["geometry"])
                except Exception:
                    pass

    return graph


def reproject_graph(graph: nx.Graph, target_crs) -> nx.Graph:
    """
    Reproject node coordinates (`x`, `y`) and edge geometries to a new CRS (in place).

    Builds GeoDataFrames for nodes and for edges that have shapely `geometry`, applies
    `GeoDataFrame.to_crs(target_crs)`, writes transformed coordinates/geometries back to the graph,
    and updates `graph.graph["crs"]` to the resulting target CRS.

    Parameters:
        graph (nx.Graph): Graph with current CRS in `graph["crs"]`; nodes carry `x`, `y`
            in that CRS, edges may carry shapely `geometry` in the same CRS.
        target_crs: Target CRS accepted by GeoPandas (EPSG int, string like `"EPSG:3857"`,
            or a `pyproj.CRS`).

    Returns:
        (nx.Graph): The same graph instance (mutated in place) with updated coordinates/geometries and CRS.

    Raises:
        ValueError: If `graph.graph["crs"]` is missing.

    Notes:
        - Only nodes with both `x` and `y` are updated.
        - Edges without shapely geometry are left unchanged.
        - If an edge `geometry` is stored as a WKT string, it is not reprojected; parse it first.
    """
    try:
        current_crs = graph.graph["crs"]
    except KeyError as exc:
        raise ValueError("Graph does not have 'crs' attribute") from exc

    nodes_items = [(n, d) for n, d in graph.nodes(data=True) if "x" in d and "y" in d]
    if nodes_items:
        node_ids = [n for n, _ in nodes_items]
        node_points = [Point(float(d["x"]), float(d["y"])) for _, d in nodes_items]
        nodes_gdf = gpd.GeoDataFrame(index=node_ids, geometry=node_points, crs=current_crs).to_crs(target_crs)
        target_crs = nodes_gdf.crs
        for nid, geom in nodes_gdf.geometry.items():
            graph.nodes[nid]["x"] = float(geom.x)
            graph.nodes[nid]["y"] = float(geom.y)

    if graph.is_multigraph():
        edge_records = [
            (u, v, k, data)
            for u, v, k, data in graph.edges(keys=True, data=True)
            if isinstance(data.get("geometry"), BaseGeometry)
        ]
        if edge_records:
            idx = [(u, v, k) for u, v, k, _ in edge_records]
            geoms = [data["geometry"] for _, _, _, data in edge_records]
            edges_gdf = gpd.GeoDataFrame(
                index=pd.MultiIndex.from_tuples(idx, names=["u", "v", "k"]), geometry=geoms, crs=current_crs
            ).to_crs(target_crs)
            target_crs = edges_gdf.crs
            for (u, v, k), geom in edges_gdf.geometry.items():
                graph.edges[u, v, k]["geometry"] = geom
    else:
        edge_records = [
            (u, v, data) for u, v, data in graph.edges(data=True) if isinstance(data.get("geometry"), BaseGeometry)
        ]
        if edge_records:
            idx = [(u, v) for u, v, _ in edge_records]
            geoms = [data["geometry"] for _, _, data in edge_records]
            edges_gdf = gpd.GeoDataFrame(
                index=pd.MultiIndex.from_tuples(idx, names=["u", "v"]), geometry=geoms, crs=current_crs
            ).to_crs(target_crs)
            target_crs = edges_gdf.crs
            for (u, v), geom in edges_gdf.geometry.items():
                graph.edges[u, v]["geometry"] = geom
    graph.graph["crs"] = target_crs
    return graph



"""
IduEdu
========

IduEdu is a Python package for the creation and manipulation of complex city networks from OpenStreetMap.

Homepage https://github.com/DDonnyy/IduEdu.
"""

from ._config import config
from ._api import *
from ._version import VERSION as __version__

from .enums import HighwayType, PublicTrasport



# pylint: disable=unused-import
from .modules.drive_walk_builders import get_drive_graph, get_walk_graph
from .modules.graph_transformers import (
    clip_nx_graph,
    gdf_to_graph,
    graph_to_gdf,
    keep_largest_strongly_connected_component,
    read_gml,
    reproject_graph,
    write_gml,
)
from .modules.intermodal_builders import get_intermodal_graph, join_pt_walk_graph
from .modules.matrix.matrix_builder import get_adj_matrix_gdf_to_gdf, get_closest_nodes
from .modules.overpass_downloaders import get_4326_boundary
from .modules.public_transport_builders import get_all_public_transport_graph, get_single_public_transport_graph



import concurrent.futures
from typing import Any

import geopandas as gpd
import networkx as nx
import pandas as pd
from shapely import LineString, MultiPolygon, Point, Polygon
from shapely.ops import substring

from iduedu import config
from iduedu.modules.drive_walk_builders import get_walk_graph
from iduedu.modules.graph_transformers import keep_largest_strongly_connected_component
from iduedu.modules.overpass_downloaders import get_4326_boundary
from iduedu.modules.public_transport_builders import get_all_public_transport_graph

logger = config.logger


def join_pt_walk_graph(
    public_transport_g: nx.Graph,
    walk_g: nx.Graph,
    max_dist=20,
    keep_largest_subgraph: bool = True,
) -> nx.Graph:
    """
    Join a public-transport graph with a pedestrian graph into a single intermodal network.

    The function relabels node ids to avoid collisions, finds PT platforms (including subway
    entries/exits), projects them onto the nearest walking edges within `max_dist`, and connects
    the two networks. When multiple platforms project to the same point on a walk edge, PT nodes
    are merged. For each platform–edge match, the walking edge is split at the projection point
    (or its endpoint is replaced by the platform node), and new directed walk edges are created
    with length/time attributes. Finally, both graphs are composed and (optionally) pruned to
    the largest strongly connected component.

    Parameters:
        public_transport_g (nx.Graph): PT graph with node attributes `x`, `y`, `type`, `route`
            and a graph CRS in `graph["crs"]`. Platform-like nodes are those with
            `type in {"platform","subway_entry_exit","subway_entry","subway_exit"}`.
        walk_g (nx.Graph): Pedestrian graph in the **same** CRS with edge `geometry` and graph
            attribute `graph["walk_speed"]` (meters/min). If `walk_speed` is missing, a default
            of 83.33 m/min is used.
        max_dist (float): Max search radius in meters to snap platforms to walk edges.
        keep_largest_subgraph (bool): If True, keep only the largest strongly connected component.

    Returns:
        (nx.Graph): Intermodal `MultiDiGraph`:
            - nodes: union of PT and Walk nodes (platforms become vertices of the walk network);
            - edges: original PT + split/updated Walk edges with `geometry`, `length_meter`,
              `time_min`, `type="walk"` (for inserted walking segments).

            Graph attrs: `graph["type"]="intermodal"`; CRS is inherited from inputs.

    Notes:
        - Node ids of PT and Walk parts are relabeled to disjoint ranges before composition,
          then relabeled densely (0..N-1) in the final graph.
        - Platforms projecting exactly to a walk edge endpoint trigger in-place node relabeling
          of that endpoint to the platform id (no edge splitting needed).
        - Duplicate projection points along the same walk edge are detected; PT platforms are merged
          and their incident PT edges are rewired onto the chosen representative.
    """

    if public_transport_g.graph["crs"] != walk_g.graph["crs"]:
        raise AttributeError(
            f"CRS mismatching, public_transport_graph crs {public_transport_g.graph['crs']}, walk_graph crs {walk_g.graph['crs']}"
        )
    logger.info("Composing intermodal graph...")

    num_nodes_g1 = len(public_transport_g.nodes)
    mapping_g1 = {node: idx for idx, node in enumerate(public_transport_g.nodes)}
    mapping_g2 = {node: idx + num_nodes_g1 for idx, node in enumerate(walk_g.nodes)}
    transport: nx.DiGraph = nx.relabel_nodes(public_transport_g, mapping_g1)
    walk: nx.MultiDiGraph = nx.relabel_nodes(walk_g, mapping_g2)

    platforms = pd.DataFrame.from_dict(dict(transport.nodes(data=True)), orient="index")
    platforms = platforms[platforms["type"].isin(["platform", "subway_entry_exit", "subway_entry", "subway_exit"])]
    platforms["geometry"] = gpd.points_from_xy(platforms["x"], platforms["y"])
    platforms = gpd.GeoDataFrame(platforms, crs=transport.graph["crs"])

    logger.debug("creating walk edges gdf")
    walk_edges = gpd.GeoDataFrame(
        nx.to_pandas_edgelist(walk.to_undirected(), source="u", target="v", edge_key="k"), crs=walk.graph["crs"]
    )
    walk_edges["edge_geometry"] = walk_edges["geometry"]

    logger.debug("joining nearest")
    projection_join = platforms.reset_index().sjoin_nearest(walk_edges, max_distance=max_dist, distance_col="dist")
    projection_join["project_dist"] = projection_join["edge_geometry"].project(projection_join["geometry"])
    projection_join["project_point"] = gpd.GeoSeries(
        projection_join["edge_geometry"].interpolate(projection_join["project_dist"]), crs=walk.graph["crs"]
    ).set_precision(1)
    projection_join["route"] = projection_join["route"].apply(lambda x: x if isinstance(x, list) else [x])

    logger.debug("searching for duplicated project points")
    projection_join = projection_join.groupby(by="project_point", as_index=False).agg(
        {
            "index": lambda x: list(dict.fromkeys(item for item in x)),
            "index_right": "first",
            "geometry": "first",
            "u": "first",
            "v": "first",
            "k": "first",
            "project_dist": "first",
            "route": lambda x: list(dict.fromkeys(item for sublist in x for item in sublist)),
        }
    )

    for ind, row in projection_join.iterrows():
        if len(row["index"]) == 1:
            projection_join.loc[ind, "index"] = row["index"][0]
        platform_in = row["index"][0]
        for platform_out in row["index"][1:]:
            if platform_out == platform_in:
                continue
            for neighbor in transport.neighbors(platform_out):
                edge_data = transport.get_edge_data(neighbor, platform_out)
                if edge_data:
                    transport.add_edge(neighbor, platform_in, **edge_data)
                edge_data = transport.get_edge_data(platform_out, neighbor)
                if edge_data:
                    transport.add_edge(platform_in, neighbor, **edge_data)
            transport.remove_node(platform_out)
        transport.nodes[platform_in]["route"] = row["route"]
        projection_join.loc[ind, "index"] = platform_in

    logger.debug("Projecting platforms to walk graph")
    points_grouped_by_edge = projection_join.groupby(by="index_right", as_index=False).agg(
        {
            "index": tuple,
            "geometry": tuple,
            "u": "first",
            "v": "first",
            "k": "first",
        }
    )
    try:
        speed = walk.graph["walk_speed"]
    except KeyError:
        logger.warning(
            "There is no walk_speed in graph, set to the default speed - 83.33 m/min"
        )  # посчитать примерную скорость по length timemin для любой эджи
        speed = 83.33

    edges_to_del = []

    # for name, row in points_grouped_by_edge.iterrows():
    for i in range(len(points_grouped_by_edge)):
        row = points_grouped_by_edge.iloc[i]
        u, v, k = row[["u", "v", "k"]]
        edge: LineString = walk.get_edge_data(u, v, k)["geometry"]
        # Если платформа одна единственная на эдж
        if len(row["index"]) == 1:
            platform_id = row["index"][0]
            dist = edge.project(row.geometry[0])
            projected_point = edge.interpolate(dist)
            if dist == 0:
                # Если платформа проецируется на начало эджа
                mapping = {u: platform_id}
                nx.relabel_nodes(walk, mapping, copy=False)
                points_grouped_by_edge.loc[points_grouped_by_edge["u"] == u, "u"] = platform_id
                points_grouped_by_edge.loc[points_grouped_by_edge["v"] == u, "v"] = platform_id
            elif dist == edge.length:
                # Если на конец
                mapping = {v: platform_id}
                nx.relabel_nodes(walk, mapping, copy=False)
                points_grouped_by_edge.loc[points_grouped_by_edge["u"] == v, "u"] = platform_id
                points_grouped_by_edge.loc[points_grouped_by_edge["v"] == v, "v"] = platform_id
            else:
                line1 = substring(edge, 0, dist)
                line2 = substring(edge, dist, edge.length)
                # Убираем старые эджи и добавляем новые, в серединке нода - наша платформа для соединения
                edges_to_del.append((u, v, k))
                if u != v:
                    edges_to_del.append((v, u, k))
                walk.add_node(platform_id, x=round(projected_point.x, 5), y=round(projected_point.y, 5))
                walk.add_edge(
                    u,
                    platform_id,
                    geometry=line1,
                    length_meter=round(line1.length, 3),
                    time_min=round(line1.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    platform_id,
                    u,
                    geometry=line1.reverse(),
                    length_meter=round(line1.length, 3),
                    time_min=round(line1.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    platform_id,
                    v,
                    geometry=line2,
                    length_meter=round(line2.length, 3),
                    time_min=round(line2.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    v,
                    platform_id,
                    geometry=line2.reverse(),
                    length_meter=round(line2.length, 3),
                    time_min=round(line2.length / speed, 3),
                    type="walk",
                )
        # Если платформм несколько на эдж
        else:
            dist_project = [
                (edge.project(platform), edge.interpolate(edge.project(platform)), ind)
                for ind, platform in zip(row["index"], row.geometry)
            ]
            dist_project.sort(key=lambda x: x[0])
            u_to_del = u
            v_to_del = v
            last_dist = 0
            last_u = u
            for dist, projected_point, cur_index in dist_project:
                if dist == 0:
                    # Если платформа проецируется на начало эджа
                    mapping = {u: cur_index}
                    u_to_del = cur_index
                    nx.relabel_nodes(walk, mapping, copy=False)
                    points_grouped_by_edge.loc[points_grouped_by_edge["u"] == u, "u"] = cur_index
                    points_grouped_by_edge.loc[points_grouped_by_edge["v"] == u, "v"] = cur_index
                elif dist == edge.length:
                    # Если на конец
                    mapping = {v: cur_index}
                    v_to_del = cur_index
                    nx.relabel_nodes(walk, mapping, copy=False)
                    points_grouped_by_edge.loc[points_grouped_by_edge["u"] == v, "u"] = cur_index
                    points_grouped_by_edge.loc[points_grouped_by_edge["v"] == v, "v"] = cur_index
                else:
                    # По очереди добавляем ноды/линии
                    line = substring(edge, last_dist, dist)
                    if isinstance(line, Point):
                        raise ValueError(f"wtf {line} cannot be linestring")
                    walk.add_node(cur_index, x=round(projected_point.x, 5), y=round(projected_point.y, 5))
                    walk.add_edge(
                        last_u,
                        cur_index,
                        geometry=line,
                        length_meter=round(line.length, 3),
                        time_min=round(line.length / speed, 3),
                        type="walk",
                    )
                    walk.add_edge(
                        cur_index,
                        last_u,
                        geometry=line.reverse(),
                        length_meter=round(line.length, 3),
                        time_min=round(line.length / speed, 3),
                        type="walk",
                    )
                    last_u = cur_index
                last_dist = dist

            # Если последняя остановка спроецировалась не на конец эджа, надо добавить остаток
            if last_dist != edge.length:
                line = substring(edge, last_dist, edge.length)
                walk.add_edge(
                    last_u,
                    v,
                    geometry=line,
                    length_meter=round(line.length, 3),
                    time_min=round(line.length / speed, 3),
                    type="walk",
                )
                walk.add_edge(
                    v,
                    last_u,
                    geometry=line.reverse(),
                    length_meter=round(line.length, 3),
                    time_min=round(line.length / speed, 3),
                    type="walk",
                )
            edges_to_del.append((u_to_del, v_to_del, k))
            if u != v:
                edges_to_del.append((v_to_del, u_to_del, k))
    walk.remove_edges_from(edges_to_del)
    logger.debug("Composing graphs")
    intermodal = nx.compose(nx.MultiDiGraph(transport), nx.MultiDiGraph(walk))
    if keep_largest_subgraph:
        intermodal = keep_largest_strongly_connected_component(intermodal)
    intermodal.remove_nodes_from([node for node, data in intermodal.nodes(data=True) if "x" not in data.keys()])
    mapping = {old_label: new_label for new_label, old_label in enumerate(intermodal.nodes())}
    logger.debug("Relabeling")
    nx.relabel_nodes(intermodal, mapping, copy=False)
    intermodal.graph["type"] = "intermodal"
    logger.debug("Done composing!")
    return intermodal


def get_intermodal_graph(
    *,
    osm_id: int | None = None,
    territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None = None,
    clip_by_territory: bool = False,
    keep_edge_geometry: bool = True,
    osm_edge_tags: list[str] | None = None,
    max_dist: float = 30,
    keep_largest_subgraph: bool = True,
    walk_kwargs: dict[str, Any] | None = None,
    pt_kwargs: dict[str, Any] | None = None,
) -> nx.Graph:
    """
    Build an intermodal (PT+walking) graph for a territory by downloading, parsing, and joining both networks.

    The function resolves a boundary polygon (by `osm_id` or `territory`), runs **in parallel**:
    1) pedestrian network construction (`get_walk_graph`),
    2) public-transport network construction for selected modes (`get_all_public_transport_graph`),
    then connects PT platforms to nearby walk edges via `join_pt_walk_graph` using a snapping radius `max_dist`.
    Edge lengths (m) and times (min) come from the underlying builders and from the walk-edge splits.

    Parameters:
        osm_id (int | None): OSM relation/area id for the territory; provide this or `territory`.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Boundary geometry in EPSG:4326.
        clip_by_territory (bool): If True, both PT and Walk graphs are clipped to the boundary.
        keep_edge_geometry (bool): If True, keep `shapely` geometries on edges for both sub-graphs.
        osm_edge_tags (list[str] | None): Subset of OSM tags to retain (forwarded to both builders).
        max_dist (float): Max distance in meters to connect PT platforms to walk edges.
        keep_largest_subgraph (bool): If True, keep only the largest strongly connected component after joining.
        walk_kwargs (dict[str, Any] | None): Extra keyword args for `get_walk_graph` (e.g., `walk_speed`,
            `simplify`, `osm_edge_tags`, `keep_largest_subgraph`, …). Defaults are sensible.
        pt_kwargs (dict[str, Any] | None): Extra keyword args for `get_all_public_transport_graph`
            (e.g., `transport_types`, `osm_edge_tags`, `keep_edge_geometry`, …).

    Returns:
        (nx.Graph): Intermodal `MultiDiGraph` combining public transport and pedestrian networks, with:
            - node attrs: `x`, `y` (local CRS), plus PT metadata for platform/station nodes where present;
            - edge attrs: `type` (e.g., "walk", PT edge types), `length_meter`, `time_min`, optional `geometry`,
              and selected OSM tags.

          Graph CRS equals the builders’ local projected CRS.

    Notes:
        - `keep_edge_geometry`, `clip_by_territory`, and `osm_edge_tags` are propagated to both builders unless
          overridden in `walk_kwargs`/`pt_kwargs`.
        - If the PT graph is empty for the area, the function returns the walking graph alone (with a warning).
        - Joining requires both sub-graphs to share the same CRS; builders derive a **local projected CRS**
          from the boundary’s extent, so lengths/times are in meters/minutes.
    """
    boundary = get_4326_boundary(osm_id=osm_id, territory=territory)

    walk_kwargs = dict(walk_kwargs or {})
    pt_kwargs = dict(pt_kwargs or {})

    for d in (walk_kwargs, pt_kwargs):
        d.setdefault("keep_edge_geometry", keep_edge_geometry)
        if osm_edge_tags is not None:
            d.setdefault("osm_edge_tags", osm_edge_tags)
        d.setdefault("clip_by_territory", clip_by_territory)

    walk_kwargs.setdefault("keep_largest_subgraph", keep_largest_subgraph)

    with concurrent.futures.ThreadPoolExecutor() as executor:
        walk_future = executor.submit(get_walk_graph, territory=boundary, **walk_kwargs)
        logger.debug("Started downloading and parsing walk graph...")

        pt_future = executor.submit(get_all_public_transport_graph, territory=boundary, **pt_kwargs)
        logger.debug("Started downloading and parsing public transport graph...")

        pt_g = pt_future.result()
        logger.debug("Public transport graph done!")

        walk_g = walk_future.result()
        logger.debug("Walk graph done!")

    if len(pt_g.nodes()) == 0:
        logger.warning("Public transport graph is empty! Returning only walk graph.")
        return walk_g

    intermodal = join_pt_walk_graph(pt_g, walk_g, max_dist=max_dist, keep_largest_subgraph=keep_largest_subgraph)
    return intermodal



import datetime as dt
import math
import time
from collections import defaultdict
from typing import Literal

import geopandas as gpd
import pandas as pd
import requests
from shapely import LineString, MultiPolygon, Polygon, unary_union
from shapely.ops import polygonize

from iduedu import config

logger = config.logger

import threading
from email.utils import parsedate_to_datetime
from time import monotonic


class RateLimiter:
    def __init__(self, min_interval: float):
        self.min_interval = float(min_interval)
        self._next_ts = 0.0
        self._cv = threading.Condition()

    def wait(self):
        start = monotonic()
        with self._cv:
            while True:
                now = monotonic()
                if now >= self._next_ts:
                    # фиксированное расписание слотов
                    self._next_ts = max(self._next_ts, now) + self.min_interval
                    slept = monotonic() - start
                    # logger.info(f"GRANT; slept={slept:.3f}s; next_slot_in={self._next_ts - now:.3f}s")
                    self._cv.notify_all()
                    return
                to_sleep = self._next_ts - now
                # logger.info(f"SLEEP for {to_sleep:.3f}s (next_ts={self._next_ts:.3f})")
                self._cv.wait(timeout=to_sleep)


OVERPASS_MIN_INTERVAL = config.overpass_min_interval
OVERPASS_RL = RateLimiter(OVERPASS_MIN_INTERVAL)


def _overpass_http(
    method: Literal["GET", "POST"], url: str, *, params=None, data=None, timeout=None
) -> requests.Response:
    OVERPASS_RL.wait()
    if method == "GET":
        return requests.get(url, params=params, timeout=timeout or config.timeout)
    return requests.post(url, data=data, timeout=timeout or config.timeout)


class RequestError(RuntimeError):
    """
    Basic error for eny requests problems
    """

    def __init__(self, message, status_code=None, reason=None, response_text=None, response_content=None):
        super().__init__(message)
        self.status_code = status_code
        self.reason = reason
        self.response_text = response_text
        self.response_content = response_content

    def __str__(self):
        if self.status_code == 400:
            return (
                f"{super().__str__()} (status: {self.status_code}, reason: {self.reason}). "
                f"Make sure provided polygon is in CRS 4326."
            )
        return f"{super().__str__()} (status: {self.status_code}, reason: {self.reason})."


def _get_overpass_pause(
    base_endpoint: str,
    recursion_pause: float = 5,
) -> float:
    url = base_endpoint[: -len("/interpreter")] + "/status"
    try:
        response = _overpass_http("GET", url, timeout=config.timeout)
        response_text = response.text
    except requests.RequestException as e:
        raise RequestError(f"Unable to reach {url}, {e}") from e

    try:
        status = response_text.split("\n")[4]
        status_first_part = status.split(" ")[0]
    except (AttributeError, IndexError, ValueError):
        raise RequestError(f"Unable to parse {url} response: {response_text}")

    try:
        _ = int(status_first_part)  # number of available slots
        pause: float = 0
    except ValueError:
        if status_first_part == "Slot":
            utc_time_str = status.split(" ")[3]
            pattern = "%Y-%m-%dT%H:%M:%SZ,"
            utc_time = dt.datetime.strptime(utc_time_str, pattern).replace(tzinfo=dt.timezone.utc)
            utc_now = dt.datetime.now(tz=dt.timezone.utc)
            seconds = int(math.ceil((utc_time - utc_now).total_seconds()))
            pause = max(seconds, 1)
        elif status_first_part == "Currently":
            time.sleep(recursion_pause)
            pause = _get_overpass_pause(base_endpoint, recursion_pause=recursion_pause)
        else:
            raise RequestError(f"Unrecognized server status: {status!r}")
    return pause


def _overpass_request(
    method: Literal["GET", "POST"],
    overpass_url: str,
    params: dict | None = None,
    data: dict | None = None,
    timeout: float | None = None,
    *,
    max_retries: int = 4,
    backoff_base: float = 1.8,
) -> requests.Response:

    last_err_text = None
    for attempt in range(max_retries + 1):
        try:
            pause = _get_overpass_pause(overpass_url)
        except RequestError as e:
            pause = 0
            logger.debug(f"Overpass /status check failed: {e}")

        if pause > 0:
            logger.warning(f"Waiting {pause} seconds for available Overpass API slot")
            time.sleep(pause)

        try:
            resp = _overpass_http(method, overpass_url, params=params, data=data, timeout=timeout or config.timeout)
        except requests.RequestException as e:
            last_err_text = str(e)
            if attempt < max_retries:
                sleep_s = min(60, backoff_base**attempt)
                logger.warning(f"Network error, retrying in {sleep_s} (attempt {attempt + 1}/{max_retries})")
                time.sleep(sleep_s)
                continue
            raise RequestError(f"Request error: {e}") from e

        if resp.status_code == 200:
            return resp

        if resp.status_code == 429:
            retry_after = resp.headers.get("Retry-After")
            if retry_after:
                try:
                    wait_s = int(retry_after)
                except ValueError:
                    try:
                        ra_dt = parsedate_to_datetime(retry_after)
                        wait_s = max(1, int((ra_dt - dt.datetime.now(tz=ra_dt.tzinfo)).total_seconds()))
                    except Exception:
                        wait_s = 5
                wait_s = max(wait_s, 1)
                logger.debug(f"HTTP 429: honoring Retry-After={wait_s}")
                time.sleep(wait_s)
            else:
                try:
                    pause = _get_overpass_pause(overpass_url)
                except RequestError:
                    pause = 0
                wait_s = max(1, pause or int(backoff_base**attempt))
                logger.debug(f"HTTP 429: waiting {wait_s} seconds before retry")
                time.sleep(wait_s)

            if attempt < max_retries:
                continue

        if resp.status_code in (502, 503, 504):
            if attempt < max_retries:
                wait_s = min(60, backoff_base**attempt)
                logger.warning(f"HTTP {resp.status_code}: retrying in {wait_s}s (attempt {attempt + 1}/{max_retries})")
                time.sleep(wait_s)
                continue
        last_err_text = resp.text
        break

    raise RequestError(
        message=f"Request failed with status code {resp.status_code}, reason: {resp.reason}",
        status_code=resp.status_code,
        reason=resp.reason,
        response_text=last_err_text,
        response_content=resp.content,
    )


def get_boundary_by_osm_id(osm_id) -> MultiPolygon | Polygon:
    overpass_query = f"""
                [out:json];
                        (
                            relation({osm_id});
                        );
                out geom;
                """
    logger.debug(f"Downloading territory bounds with osm_id <{osm_id}> ...")
    resp = _overpass_request(
        method="GET",
        overpass_url=config.overpass_url,
        params={"data": overpass_query},
    )
    json_result = resp.json()
    geoms = []
    for element in json_result["elements"]:
        geometries_inners = []
        geometries_outers = []
        for member in element["members"]:
            if "geometry" in member:
                if member["role"] == "outer":
                    geometries_outers.append(
                        LineString([(coords["lon"], coords["lat"]) for coords in member["geometry"]])
                    )
                if member["role"] == "inner":
                    geometries_inners.append(
                        LineString([(coords["lon"], coords["lat"]) for coords in member["geometry"]])
                    )
        outer_poly = unary_union(list(polygonize(geometries_outers)))
        inner_poly = unary_union(list(polygonize(geometries_inners)))
        geoms.append(outer_poly.difference(inner_poly))
    return unary_union(geoms)


def get_4326_boundary(
    *, osm_id: int | None = None, territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None = None
) -> Polygon:
    """
    Normalize a territory boundary to a single EPSG:4326 Polygon.

    Accepts either an `osm_id` (relation id) to fetch the boundary from OSM, a direct
    `Polygon`/`MultiPolygon`, or a `GeoDataFrame`. Returns a single `Polygon` in lon/lat.
    For `MultiPolygon`, the function returns its **convex hull** as a Polygon.

    Parameters:
        osm_id (int | None): OSM relation id. If provided, boundary is fetched via Overpass.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Existing boundary geometry
            or a GeoDataFrame containing it. If GeoDataFrame is given, it is reprojected to 4326
            and unioned (`.union_all()`).

    Returns:
        (shapely.Polygon): Boundary polygon in EPSG:4326.

    Notes:
        - Input `Polygon` is returned as-is (assumed already in EPSG:4326 by caller).
        - For `MultiPolygon`, a **convex hull** is returned (may slightly expand the area and
          fill gaps between parts).
        - For `GeoDataFrame`, the geometry is first reprojected to 4326 and dissolved via
          `.union_all()`; the result is then normalized to a `Polygon` (convex hull if needed).

    Examples:
        >>> get_4326_boundary(osm_id=1114252)             # fetch by OSM id
        >>> get_4326_boundary(territory=poly4326)         # keep polygon
        >>> get_4326_boundary(territory=multi_poly_4326)  # convex hull of multipart
        >>> get_4326_boundary(territory=territory_gdf)    # GDF -> to_crs(4326) -> union_all -> Polygon
    """
    if osm_id:
        territory = get_boundary_by_osm_id(osm_id)

    if isinstance(territory, Polygon):
        return territory

    if isinstance(territory, MultiPolygon):
        return Polygon(territory.convex_hull)

    if isinstance(territory, gpd.GeoDataFrame):
        return get_4326_boundary(territory=territory.to_crs(4326).union_all())

    raise ValueError("Either osm_id or polygon must be specified")


def _poly_to_overpass(poly: Polygon) -> str:
    return " ".join(f"{y} {x}" for x, y in poly.exterior.coords[:-1])


def get_routes_by_poly(polygon: Polygon, public_transport_type: str) -> pd.DataFrame:
    if public_transport_type == "subway":
        return get_subway_routes_by_poly(polygon)
    polygon_coords = _poly_to_overpass(polygon)
    overpass_query = f"""
        [out:json][timeout:{config.timeout}];
                (
                    relation(poly:\"{polygon_coords}\")[ 'route' = '{public_transport_type}' ];
                );
        out geom;
        """
    logger.debug(f"Downloading routes from OSM with type <{public_transport_type}> ...")
    resp = _overpass_request(
        method="POST",
        overpass_url=config.overpass_url,
        data={"data": overpass_query},
    )
    json_result = resp.json()["elements"]
    data = pd.DataFrame(json_result)
    data["transport_type"] = public_transport_type
    return data


def get_subway_routes_by_poly(polygon: Polygon) -> pd.DataFrame:
    polygon_coords = _poly_to_overpass(polygon)
    overpass_query = f"""
        [out:json][timeout:500];
            rel(poly:"{polygon_coords}")["route"="subway"]->.routes;
            node(r.routes)-> .route_nodes;
            rel(bn.route_nodes)->.stop_areas;
            rel(br.stop_areas)["public_transport"="stop_area_group"]["type"="public_transport"]->.groups;
            nwr(r.stop_areas)["public_transport"="station"]->.stations;
            .stop_areas     out geom qt;
            .groups         out body qt;
            .stations       out tags qt;
        """

    logger.debug(f"Downloading subway routes data from OSM ...")
    resp = _overpass_request(
        method="POST",
        overpass_url=config.overpass_url,
        data={"data": overpass_query},
    )
    json_result = resp.json()["elements"]
    if len(json_result) == 0:
        return pd.DataFrame()

    for e in json_result:
        tags = e.get("tags") or {}
        etype = e.get("type")

        e["is_stop_area"] = etype == "relation" and tags.get("public_transport") == "stop_area"
        e["is_stop_area_group"] = (
            etype == "relation"
            and tags.get("public_transport") == "stop_area_group"
            and tags.get("type") == "public_transport"
        )
        e["is_station"] = tags.get("public_transport") == "station"

    data = pd.DataFrame(json_result)
    data["transport_type"] = "subway"
    return data


def get_network_by_filters(polygon: Polygon, way_filter: str) -> pd.DataFrame:
    polygon_coords = _poly_to_overpass(polygon)
    overpass_query = f"""
        [out:json][timeout:{config.timeout}];
            (way{way_filter}(poly:\"{polygon_coords}\"););
        out geom;
        """
    logger.debug(f"Downloading network from OSM with filters <{way_filter}> ...")
    resp = _overpass_request(
        method="POST",
        overpass_url=config.overpass_url,
        data={"data": overpass_query},
    )

    json_result = resp.json()["elements"]
    return pd.DataFrame(json_result)


def fetch_member_tags(members_missing, chunk_size=2000):
    """
    members_missing: iterable of dicts  {'type': 'node|way|relation', 'ref': int}
    :returns {("node", 123): {tags...}, ("way", 456): {tags...}, ...}
    """
    ids = defaultdict(list)
    for m in members_missing:
        t = m["type"]
        ids[t].append(int(m["ref"]))

    for k in list(ids.keys()):
        ids[k] = sorted(set(ids[k]))

    result = {}

    def _build_query(sub_ids: dict) -> str:
        parts = []
        if sub_ids.get("node"):
            parts.append(f'node(id:{",".join(map(str, sub_ids["node"]))});')
        if sub_ids.get("way"):
            parts.append(f'way(id:{",".join(map(str, sub_ids["way"]))});')
        if sub_ids.get("relation"):
            parts.append(f'rel(id:{",".join(map(str, sub_ids["relation"]))});')
        body = "\n".join(parts)
        return f"[out:json][timeout:{config.timeout}];\n(\n{body}\n);\nout tags center qt;"

    def _yield_chunks(type_key):
        arr = ids.get(type_key, [])
        for i in range(0, len(arr), chunk_size):
            yield {type_key: arr[i : i + chunk_size]}

    chunks = []
    for tk in ("node", "way", "relation"):
        chunks.extend(_yield_chunks(tk))

    for sub in chunks:
        q = _build_query(sub)
        resp = _overpass_request(method="POST", overpass_url=config.overpass_url, data={"data": q})
        els = resp.json().get("elements", [])
        for e in els:
            key = (e["type"], int(e["id"]))
            result[key] = e.get("tags", {}) or {}
            if e["type"] != "node":
                if "center" in e:
                    result[key]["__center__"] = (e["center"]["lon"], e["center"]["lat"])

    return result



from typing import Literal

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from pyproj import CRS
from shapely import LineString, MultiLineString, Polygon, line_merge
from shapely.geometry.multipolygon import MultiPolygon

from iduedu import config
from iduedu.enums.highway_enums import HighwayType
from iduedu.enums.network_enums import Network
from iduedu.modules.graph_transformers import estimate_crs_for_bounds, keep_largest_strongly_connected_component
from iduedu.modules.overpass_downloaders import get_4326_boundary, get_network_by_filters

logger = config.logger


def _get_highway_properties(highway) -> tuple[str, float]:
    """
    Map OSM `highway` class(es) to a regulatory category and a default speed.

    Accepts either a single string or a list of classes. When multiple classes are present,
    picks the lowest (most permissive) regulatory status and the minimum default speed among them.

    Parameters:
        highway (str | list[str]): OSM highway class or list of classes (e.g., "primary", "residential", ...).

    Returns:
        (tuple[str, float]): A pair `(category, maxspeed_mpm)` where:
            - `category` (str): regulatory class label, e.g., "local" | "regional" | "federal".
            - `maxspeed_mpm` (float): default speed for that class in **meters per minute**.

    Notes:
        If the class is unknown, returns defaults for `UNCLASSIFIED`.
    """
    default = (HighwayType.UNCLASSIFIED.reg_status, HighwayType.UNCLASSIFIED.max_speed)

    if not highway:
        return default

    highway_list = highway if isinstance(highway, list) else [highway]
    reg_values, speed_values = [], []
    for ht in highway_list:
        try:
            enum_value = HighwayType[ht.upper()]
        except KeyError:
            continue
        reg_values.append(enum_value.reg_status)
        speed_values.append(enum_value.max_speed)

    if not reg_values or not speed_values:
        return default

    rank = {"local": 0, "regional": 1, "federal": 2}
    lowest_reg = min(reg_values, key=lambda s: rank.get(s, 999))

    min_speed = min(speed_values) if speed_values else default[1]

    return lowest_reg, min_speed


def _build_edges_from_overpass(polygon: Polygon, way_filter: str, simplify: bool = True) -> tuple[GeoDataFrame, CRS]:
    """
    Download OSM ways by filter, segment into edges, and project to a local CRS.

    The function queries the ways within `polygon` using an Overpass `way_filter`, converts each way
    into consecutive line segments (one edge per segment), and projects the result to a local metric CRS.
    If `simplify=True`, contiguous segments are merged (`line_merge`), and OSM attributes are transferred
    to merged lines via midpoint nearest join.

    Parameters:
        polygon (Polygon): Boundary polygon in EPSG:4326 used to query OSM data.
        way_filter (str): Overpass filter applied to ways (e.g., `[ "highway" ~ "motorway|trunk|primary|..." ]`).
        simplify (bool): If True, merges contiguous segments into longer edges and reattaches attributes
            using a nearest midpoint join.

    Returns:
        (tuple[gpd.GeoDataFrame, CRS]): A pair `(edges, local_crs)` where:
            - `edges` (GeoDataFrame): Line features in local CRS with columns:
                `geometry` (LineString), `way_idx` (source way index), `id` (OSM way id), `tags` (dict).
            - `local_crs` (pyproj.CRS): Estimated local projected CRS suitable for metric length computations.

    Notes:
        Attributes on merged edges are inferred from the nearest original segment around the midpoint,
        which may drop or aggregate original per-segment variability.
    """
    data = get_network_by_filters(polygon, way_filter)
    ways = data[data["type"] == "way"].copy()

    # Собираем координаты каждой линии (lon, lat)
    coords_list = [np.asarray([(p["lon"], p["lat"]) for p in pts], dtype="f8") for pts in ways["geometry"].values]

    # сегментация на отрезки
    starts = np.concatenate([a[:-1] for a in coords_list], axis=0)
    ends = np.concatenate([a[1:] for a in coords_list], axis=0)

    lengths = np.array([a.shape[0] for a in coords_list], dtype=int)
    seg_counts = np.maximum(lengths - 1, 0)
    way_idx = np.repeat(ways.index.values, seg_counts)

    geoms = [LineString([tuple(s), tuple(e)]) for s, e in zip(starts, ends)]

    # локальная проекция
    local_crs = estimate_crs_for_bounds(*polygon.bounds)

    edges = gpd.GeoDataFrame({"way_idx": way_idx}, geometry=geoms, crs=4326).to_crs(local_crs)
    edges = edges.join(ways[["id", "tags"]], on="way_idx")

    if simplify:
        # сшиваем ребра и переносим атрибуты через midpoints -> nearest
        merged = list(
            line_merge(MultiLineString(edges.geometry.to_list()), directed=True).geoms
        )  # TODO Можно быстрее, если разьединить по тегам и распараллелить, тоже самое в DRIVE
        lines = gpd.GeoDataFrame(geometry=merged, crs=local_crs)

        mid = lines.copy()
        mid.geometry = mid.interpolate(lines.length / 2)

        # TODO Вот тут подумать надо, теряются аттрибуты, надо ли складывать их "по умному", в листы.
        joined = gpd.sjoin_nearest(mid[["geometry"]], edges, how="left", max_distance=1)
        joined = joined.reset_index().drop_duplicates(subset="index").set_index("index")
        lines = lines.join(joined[["tags", "id", "way_idx"]])

        edges = lines

    return edges, local_crs


def get_drive_graph(
    *,
    osm_id: int | None = None,
    territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None = None,
    simplify: bool = True,
    add_road_category: bool = True,
    clip_by_territory: bool = False,
    keep_largest_subgraph: bool = True,
    network_type: Literal["drive", "drive_service", "custom"] = "drive",
    custom_filter: str | None = None,
    osm_edge_tags: list[str] | None = None,  # overrides default tags
    keep_edge_geometry: bool = True,
) -> nx.MultiDiGraph:
    """
    Build a drivable road network (nx.MultiDiGraph) from OpenStreetMap within a given territory.

    The function downloads OSM ways via Overpass, segments them into directed edges, optionally merges
    contiguous segments, duplicates two-way streets in reverse, and computes per-edge length (meters)
    and travel time (minutes). Node coordinates are unique line endpoints in a local projected CRS.
    Edge attributes can include selected OSM tags and a derived road category/speed.

    Parameters:
        osm_id (int | None): OSM relation/area ID of the territory boundary. Provide this or `territory`.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Boundary geometry in EPSG:4326 or GeoDataFrame.
            Used when `osm_id` is not given.
        simplify (bool): If True, merges contiguous collinear segments and transfers attributes
            back to merged lines using nearest midpoints. If False, keeps raw per-segment edges.
        add_road_category (bool): If True, adds a derived `category` (e.g., local/regional/federal) and a default
            speed (`maxspeed_mpm`) inferred from the OSM `highway` class.
        clip_by_territory (bool): If True, clips edges by the exact boundary geometry before graph construction.
        keep_largest_subgraph (bool): If True, returns only the largest strongly connected component.
        network_type (Literal["drive","drive_service","custom"]): Preset of Overpass filters to select drivable ways.
            Use "custom" together with `custom_filter` to pass your own Overpass `way` filter.
        custom_filter (str | None): Custom Overpass filter (e.g., `["highway"~"motorway|trunk|…"]`) used when
            `network_type="custom"`.
        osm_edge_tags (list[str] | None): Which OSM tags to retain on edges. Overrides defaults from config.
            The tags `oneway`, `maxspeed`, and `highway` are always added.
        keep_edge_geometry (bool): If True, stores shapely geometries on edges (`geometry` attribute).

    Returns:
        (nx.MultiDiGraph): Directed multigraph of the road network. Each edge carries:
            - `geometry` (if `keep_edge_geometry=True`) in local CRS,
            - `length_meter` (float), `time_min` (float),
            - `type="drive"`,
            - selected OSM tags (incl. `highway`, `maxspeed`, `oneway`, optional `category`, etc.).

            Graph-level attributes: `graph["crs"]` (local projected CRS), `graph["type"]` (network_type).

    Raises:
        ValueError: If `network_type` is unknown, or `network_type="custom"` without `custom_filter`.
    """
    polygon4326 = get_4326_boundary(osm_id=osm_id, territory=territory)

    filters = {
        "drive": Network.DRIVE.filter,
        "drive_service": Network.DRIVE_SERVICE.filter,
        "custom": custom_filter,
    }
    try:
        road_filter = filters[network_type]
    except KeyError:
        raise ValueError(f"Unknown road_type: {network_type!r}")
    if network_type == "custom" and road_filter is None:
        raise ValueError("For road_type='custom' you must provide custom_filter")

    logger.info("Downloading drive network via Overpass ...")
    edges, local_crs = _build_edges_from_overpass(polygon4326, road_filter, simplify=simplify)

    if osm_edge_tags is None:
        needed_tags = set(config.drive_useful_edges_attr)
    else:
        needed_tags = set(osm_edge_tags)

    tags_to_retrieve = set(needed_tags) | {"oneway", "maxspeed", "highway"}

    tags_df = pd.DataFrame.from_records(
        ({k: v for k, v in d.items() if k in tags_to_retrieve} for d in edges["tags"]),
        index=edges.index,
    )
    edges = edges.join(tags_df)

    if clip_by_territory:
        clip_poly_gdf = gpd.GeoDataFrame(geometry=[polygon4326], crs=4326).to_crs(local_crs)
        edges = edges.clip(clip_poly_gdf, keep_geom_type=True)

    # двусторонние — дублируем с реверсом
    two_way = edges[edges["oneway"] != "yes"].copy()
    two_way.geometry = two_way.geometry.reverse()
    edges = pd.concat([edges, two_way], ignore_index=True)

    coords = edges.geometry.get_coordinates().to_numpy()
    counts = edges.geometry.count_coordinates()
    cuts = np.cumsum(counts)
    first_idx = np.r_[0, cuts[:-1]]
    last_idx = cuts - 1
    starts = coords[first_idx]
    ends = coords[last_idx]
    edges["start"] = list(map(tuple, starts))
    edges["end"] = list(map(tuple, ends))

    all_endpoints = pd.Index(edges["start"]).append(pd.Index(edges["end"]))
    labels, uniques = pd.factorize(all_endpoints)
    n = len(edges)
    u = labels[:n]
    v = labels[n:]
    edges["u"] = u
    edges["v"] = v

    edges[["category", "maxspeed_mpm"]] = edges["highway"].apply(lambda h: pd.Series(_get_highway_properties(h)))

    maxspeed_osm_mpm = (pd.to_numeric(edges["maxspeed"], errors="coerce") * 1000.0 / 60.0).round(3)

    edges["speed_mpm"] = maxspeed_osm_mpm.fillna(edges["maxspeed_mpm"])

    edges["length_meter"] = edges.geometry.length.round(3)
    edges["time_min"] = (edges["length_meter"] / edges["speed_mpm"]).round(3)
    edges["type"] = "drive"

    graph = nx.MultiDiGraph()

    graph.add_nodes_from((i, {"x": float(x), "y": float(y)}) for i, (x, y) in enumerate(uniques))

    if add_road_category:
        needed_tags |= {"category"}
    if keep_edge_geometry:
        needed_tags |= {"geometry"}
    needed_tags |= {"length_meter", "time_min", "type"}
    edge_attr_cols = list(tag for tag in needed_tags if tag in edges.columns)
    attrs_iter = edges[edge_attr_cols].to_dict("records")
    graph.add_edges_from((int(uu), int(vv), d) for uu, vv, d in zip(u, v, attrs_iter))

    if keep_largest_subgraph:
        graph = keep_largest_strongly_connected_component(graph)

    mapping = {old: new for new, old in enumerate(graph.nodes())}
    graph = nx.relabel_nodes(graph, mapping)
    graph.graph["crs"] = local_crs
    graph.graph["type"] = network_type
    logger.debug("Drive graph built.")
    return graph


def get_walk_graph(
    *,
    osm_id: int | None = None,
    territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None = None,
    simplify: bool = True,
    clip_by_territory: bool = False,
    keep_largest_subgraph: bool = True,
    walk_speed: float = 5 * 1000 / 60,  # m/min
    network_type: Literal["walk", "custom"] = "walk",
    custom_filter: str | None = None,
    osm_edge_tags: list[str] | None = None,  # overrides default tags
    keep_edge_geometry: bool = True,
) -> nx.MultiDiGraph:
    """
    Build a pedestrian network (nx.MultiDiGraph) from OpenStreetMap within a given territory.

    The function fetches OSM ways via Overpass using a walking filter, splits each way into directed
    line segments, duplicates all segments in reverse, and computes per-edge length (meters) and traversal time
    (minutes) using a given walking speed (m/min). Node coordinates are unique segment endpoints in a local
    projected CRS. Selected OSM tags can be attached to edges.

    Parameters:
        osm_id (int | None): OSM relation/area ID for the boundary. Provide this or `territory`.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Boundary geometry (EPSG:4326) or a GeoDataFrame
            to define the area when `osm_id` is not given.
        simplify (bool): If True, merges contiguous segments (via internal line merging) and transfers attributes
            back to merged lines using nearest midpoints; if False, keeps raw per-segment edges.
        clip_by_territory (bool): If True, clips edges by the provided boundary before graph construction.
        keep_largest_subgraph (bool): If True, retains only the largest strongly connected component.
        walk_speed (float): Walking speed in meters per minute used to compute `time_min` for each edge.
        network_type (Literal["walk","custom"]): Preset of Overpass filters. Use "custom" together with `custom_filter`
            to pass your own way filter.
        custom_filter (str | None): Custom Overpass filter string used when `network_type="custom"`.
        osm_edge_tags (list[str] | None): List of OSM edge tags to retain (overrides defaults). Only these keys
            are joined from element tags.
        keep_edge_geometry (bool): If True, stores shapely `geometry` on edges in the local projected CRS.

    Returns:
        (nx.MultiDiGraph): Directed multigraph of the walking network. Each edge carries:
            - `geometry` (if `keep_edge_geometry=True`), local CRS,
            - `length_meter` (float), `time_min` (float),
            - `type="walk"`,
            - selected OSM tags (as requested).

            Graph attributes include: `graph["crs"]` (local projected CRS),
            `graph["walk_speed"]` (float), and `graph["type"]` (network_type).

    Raises:
        ValueError: If `network_type` is unknown, or `network_type="custom"` without `custom_filter`.

    Notes:
        All walking edges are treated as bidirectional by duplicating geometries in reverse; u/v nodes
        are assigned by factorizing unique segment endpoints. Lengths are measured in meters in a local
        projected CRS estimated from the territory bounds.
    """
    polygon4326 = get_4326_boundary(osm_id=osm_id, territory=territory)

    filters = {
        "walk": Network.WALK.filter,
        "custom": custom_filter,
    }
    try:
        road_filter = filters[network_type]
    except KeyError:
        raise ValueError(f"Unknown road_type: {network_type!r}")
    if network_type == "custom" and road_filter is None:
        raise ValueError("For road_type='custom' you must provide custom_filter")

    logger.info("Downloading walk network via Overpass ...")
    edges, local_crs = _build_edges_from_overpass(polygon4326, road_filter, simplify=simplify)

    if osm_edge_tags is None:
        needed_tags = set(config.walk_useful_edges_attr)
    else:
        needed_tags = set(osm_edge_tags)

    tags_df = pd.DataFrame.from_records(
        ({k: v for k, v in d.items() if k in needed_tags} for d in edges["tags"]),
        index=edges.index,
    )
    edges = edges.join(tags_df)

    if clip_by_territory:
        clip_poly_gdf = gpd.GeoDataFrame(geometry=[polygon4326], crs=4326).to_crs(local_crs)
        edges = edges.clip(clip_poly_gdf, keep_geom_type=True)

    two_way = edges.copy()
    two_way.geometry = two_way.geometry.reverse()
    edges = pd.concat([edges, two_way], ignore_index=True)

    # u/v и узлы
    coords = edges.geometry.get_coordinates().to_numpy()
    counts = edges.geometry.count_coordinates()
    cuts = np.cumsum(counts)
    first_idx = np.r_[0, cuts[:-1]]
    last_idx = cuts - 1
    starts = coords[first_idx]
    ends = coords[last_idx]
    edges["start"] = list(map(tuple, starts))
    edges["end"] = list(map(tuple, ends))

    all_endpoints = pd.Index(edges["start"]).append(pd.Index(edges["end"]))
    labels, uniques = pd.factorize(all_endpoints)
    n = len(edges)
    u = labels[:n]
    v = labels[n:]

    edges["length_meter"] = edges.geometry.length.round(3)
    edges["time_min"] = (edges["length_meter"] / float(walk_speed)).round(3)
    edges["type"] = "walk"

    # Сборка графа
    graph = nx.MultiDiGraph()
    graph.add_nodes_from((i, {"x": float(x), "y": float(y)}) for i, (x, y) in enumerate(uniques))

    edge_attrs = set(needed_tags)
    if keep_edge_geometry:
        edge_attrs |= {"geometry"}
    edge_attrs |= {"length_meter", "time_min", "type"}

    attrs_iter = edges[list(edge_attrs)].to_dict("records")
    graph.add_edges_from((int(uu), int(vv), d) for uu, vv, d in zip(u, v, attrs_iter))

    if keep_largest_subgraph:
        graph = keep_largest_strongly_connected_component(graph)

    mapping = {old: new for new, old in enumerate(graph.nodes())}
    graph = nx.relabel_nodes(graph, mapping)
    graph.graph["crs"] = local_crs
    graph.graph["walk_speed"] = float(walk_speed)
    graph.graph["type"] = network_type
    logger.debug("Walk graph built.")
    return graph



from heapq import heappop, heappush
from typing import Any, Iterable, Literal

import geopandas as gpd
import networkx as nx
import numba as nb
import numpy as np
import pandas as pd
from numpy.core.multiarray import ndarray
from pyproj import CRS
from pyproj.exceptions import CRSError
from scipy.spatial import KDTree

from iduedu import config
from iduedu.modules.graph_transformers import keep_largest_strongly_connected_component
from iduedu.modules.matrix.numba_csr_matrix import UI32CSRMatrix

logger = config.logger


@nb.njit(
    [
        nb.types.Array(nb.int32, 2, "C")(
            UI32CSRMatrix.class_type.instance_type,
            nb.types.Array(nb.int32, 1, "C"),
            nb.types.Array(nb.int32, 1, "C"),
            nb.int32,
        )
    ],
    cache=True,
    parallel=True,
)
def dijkstra_numba_parallel(
    numba_matrix: UI32CSRMatrix, sources: np.array, targets: np.array, cutoff: np.int32
):  # pragma: no cover
    """
    Parallel multi-source Dijkstra on a CSR matrix (Numba-accelerated).

    Runs Dijkstra from each `source` over a directed graph encoded as a uint32-weight CSR
    (`UI32CSRMatrix`). Distances are accumulated as **int32** in the same units as the matrix
    edge weights (this implementation expects weights pre-scaled by ×100; see helpers below).
    For each source, distances to `targets` are returned; unreachable entries are `-1`.

    Parameters:
        numba_matrix (UI32CSRMatrix): Directed graph in CSR form with non-negative integer weights.
            Attributes used: `.tot_rows`, `.get_row(i)` (weights), `.get_nnz(i)` (neighbor indices).
        sources (np.ndarray[int32]): Node indices to start from.
        targets (np.ndarray[int32]): Node indices to read distances for.
        cutoff (np.int32): Maximum distance; paths exceeding this are ignored.

    Returns:
        (np.ndarray[int32]): Matrix of shape (len(sources), len(targets)); `-1` marks “no path”.
    """
    distance_matrix = np.full((len(sources), len(targets)), -1, dtype=np.int32)

    for i in nb.prange(len(sources)):
        source = sources[i]
        distances = np.full(numba_matrix.tot_rows, -1, dtype=np.int32)
        seen = np.copy(distances)
        fringe = []
        seen[source] = np.int32(0)
        fringe.append((np.int32(0), np.int32(source)))
        while fringe:
            d, v = heappop(fringe)
            if distances[v] != -1:
                continue
            distances[v] = d
            for u, weight in zip(numba_matrix.get_nnz(v), numba_matrix.get_row(v)):
                uv_dist = np.int32(d + weight)
                if uv_dist > cutoff:
                    continue
                if seen[u] == -1 or uv_dist < seen[u]:
                    seen[u] = uv_dist
                    heappush(fringe, (uv_dist, np.int32(u)))

        distance_matrix[i, :] = distances[targets]
    return distance_matrix


def get_closest_nodes(
    gdf_from: gpd.GeoDataFrame, to_nx_graph: nx.Graph
) -> tuple[list[Any], float | ndarray | Iterable | int]:
    """
    Find the nearest graph node for each geometry in a GeoDataFrame.

    Reprojects `gdf_from` to the graph CRS if needed, then builds a KD-tree over node
    coordinates (`x`, `y`) and queries nearest nodes for point representatives of the
    input geometries (handles points/lines/polygons via `representative_point()`).

    Parameters:
        gdf_from (gpd.GeoDataFrame): Input geometries to snap (any geometry type).
        to_nx_graph (nx.Graph): Graph with node attributes `x`, `y` and `graph["crs"]` set.

    Returns:
        (list, np.ndarray): Tuple of:
            - `nearest_nodes`: list of node IDs (in the graph index space),
            - `distances`: NumPy array of Euclidean distances in the graph CRS units.
    """
    try:
        graph_crs = CRS.from_epsg(to_nx_graph.graph["crs"])
    except CRSError:
        graph_crs = to_nx_graph.graph["crs"]

    if gdf_from.crs != graph_crs:
        gdf_from = gdf_from.to_crs(graph_crs)

    nodes_with_data = list(to_nx_graph.nodes(data=True))
    coordinates = [(data["x"], data["y"]) for node, data in nodes_with_data]
    tree = KDTree(coordinates)
    target_coord = [(p.x, p.y) for p in gdf_from.representative_point()]
    distances, indices = tree.query(target_coord)
    nearest_nodes = [nodes_with_data[idx][0] for idx in indices]

    return nearest_nodes, distances


def get_adj_matrix_gdf_to_gdf(
    gdf_from: gpd.GeoDataFrame,
    gdf_to: gpd.GeoDataFrame,
    nx_graph: nx.Graph,
    weight: Literal["length_meter", "time_min"] = "length_meter",
    dtype: np.dtype = np.float16,
    add_dist_tofrom_node=True,
    threshold: int = None,
    max_workers: int = None,
) -> pd.DataFrame:
    """
    Compute a shortest-path matrix between two GeoDataFrames over a weighted NetworkX graph.

    Each origin (row of `gdf_from`) and destination (column of `gdf_to`) is snapped to its nearest
    graph node (KD-tree), then distances are computed with a batched, Numba-parallel Dijkstra on
    the graph’s CSR matrix. Result units follow `weight`:
    - `"length_meter"` → meters,
    - `"time_min"` → minutes.

    Parameters:
        gdf_from (gpd.GeoDataFrame): Origin geometries (any types; snapped via representative points).
        gdf_to (gpd.GeoDataFrame): Destination geometries.
        nx_graph (nx.Graph): Graph with `graph["crs"]` and per-edge `weight` attribute present.
        weight ({"length_meter", "time_min"}): Edge attribute to minimize.
        dtype (np.dtype): Output matrix dtype (default `np.float16` to save memory).
        add_dist_tofrom_node (bool): If True, adds straight-line distances from origin geometry → its snap node
            and from snap node → destination geometry:
            - if `weight="length_meter"`: meters are added,
            - if `weight="time_min"`: meters converted to minutes assuming 5 km/h (~83.33 m/min).
        threshold (int | None): Optional max path threshold in `weight` units; longer paths are treated as missing.
            Internally quantized by ×100 for integer CSR; `None` ⇒ no cutoff.
        max_workers (int | None): If set, limits Numba thread count (`nb.set_num_threads(max_workers)`).

    Returns:
        (pd.DataFrame): Matrix with index = `gdf_from.index`, columns = `gdf_to.index`.
            Values are shortest-path distances; unreachable pairs are `np.inf`.

    Raises:
        ValueError: If the graph lacks `graph["crs"]`.
        CRSError: If the graph CRS is invalid or reprojection fails.

    Notes:
        - The graph is first pruned to its largest strongly connected component to avoid spurious `∞`.
        - For performance, if `len(gdf_from) > len(gdf_to)`, the computation is performed transposed and flipped back.
        - Internally, edge weights are converted to **uint32** by multiplying by 100; results are divided by 100
          before returning.
    """
    try:
        local_crs = nx_graph.graph["crs"]
    except KeyError as exc:
        raise ValueError("Graph does not have crs attribute") from exc

    try:
        gdf_from = gdf_from.to_crs(nx_graph.graph["crs"])
        gdf_to = gdf_to.to_crs(nx_graph.graph["crs"])
    except CRSError as e:
        raise CRSError(f"Graph crs ({local_crs}) has invalid format.") from e

    nx_graph = keep_largest_strongly_connected_component(nx_graph)

    logger.debug("Preparing graph sparse matrix")
    transposed = False

    if gdf_from.equals(gdf_to):
        closest_nodes_from, dist_from = get_closest_nodes(gdf_from, nx_graph)
        closest_nodes_to, dist_to = closest_nodes_from, dist_from
        sparse_row_scipy = _get_sparse_row(nx_graph, weight)
    else:
        dif = len(gdf_from) - len(gdf_to)
        if dif <= 0:  # straight
            closest_nodes_from, dist_from = get_closest_nodes(gdf_from, nx_graph)
            closest_nodes_to, dist_to = get_closest_nodes(gdf_to, nx_graph)
            sparse_row_scipy = _get_sparse_row(nx_graph, weight)

        else:  # reversed
            transposed = True
            closest_nodes_from, dist_from = get_closest_nodes(gdf_to, nx_graph)
            closest_nodes_to, dist_to = get_closest_nodes(gdf_from, nx_graph)
            sparse_row_scipy = _get_sparse_row(nx_graph, weight).transpose().tocsc().tocsr()

    csr_matrix = UI32CSRMatrix(*_get_numba_matrix_attr(sparse_row_scipy))
    del sparse_row_scipy

    if max_workers is not None:
        nb.set_num_threads(max_workers)

    # *** Warm-up ***
    dijkstra_numba_parallel(
        numba_matrix=csr_matrix,
        sources=np.array(closest_nodes_to[:1], dtype=np.int32),
        targets=np.array(closest_nodes_from[:1], dtype=np.int32),
        cutoff=np.int32(0),
    )

    if threshold is None:
        threshold = np.iinfo(np.int32).max
    else:
        threshold = threshold * 100

    logger.debug("Starting the gdf-to-gdf matrix calculation")

    adj_matrix = dijkstra_numba_parallel(
        numba_matrix=csr_matrix,
        sources=np.array(closest_nodes_from, dtype=np.int32),
        targets=np.array(closest_nodes_to, dtype=np.int32),
        cutoff=np.int32(threshold),
    )
    if transposed:
        adj_matrix = adj_matrix.transpose()

    adj_matrix = pd.DataFrame(adj_matrix / 100, columns=gdf_to.index, index=gdf_from.index, dtype=dtype)
    if add_dist_tofrom_node:
        if weight == "time_min":
            speed = 5 * 1000 / 60
            dist_to = dist_to / speed
            dist_from = dist_from / speed
        dist_from_matrix = np.array(dist_from)[:, np.newaxis]
        dist_to_matrix = np.array(dist_to)[np.newaxis, :]
        mask = adj_matrix > 0
        additional_matrix = (dist_from_matrix + dist_to_matrix).astype(dtype)
        if transposed:
            additional_matrix = additional_matrix.transpose()
        adj_matrix[mask] += additional_matrix
    return adj_matrix.where(adj_matrix >= 0, np.inf)


def _get_sparse_row(nx_graph, weight):
    """
    Convert a (possibly multi-)graph to a SciPy CSR matrix of edge weights (uint32, ×100 scaling).

    If the graph is a Multi(Graph|DiGraph), it is first coerced to a simple Graph/DiGraph
    (parallel edges are **collapsed and their attributes lost**, keeping one arbitrary edge).
    Then `nx.to_scipy_sparse_array(..., weight=weight)` produces a CSR array; its data are
    multiplied by 100 and cast to `uint32` for integer-only NumPy/Numba arithmetic.

    Parameters:
        nx_graph (nx.Graph | nx.DiGraph | nx.MultiGraph | nx.MultiDiGraph): Input network.
        weight (str): Name of the edge attribute to use as weight (must be numeric and non-negative).

    Returns:
        (scipy.sparse.csr_array): CSR matrix with `uint32` data scaled by ×100.

    Warnings:
        Collapsing a multigraph to a simple graph discards multi-edge multiplicity and may change
        effective weights if parallel edges existed.
    """
    if nx_graph.is_multigraph():
        if nx_graph.is_directed():
            nx_graph = nx.DiGraph(nx_graph)
        else:
            nx_graph = nx.Graph(nx_graph)
    sparse_row_scipy = nx.to_scipy_sparse_array(nx_graph, weight=weight)
    sparse_row_scipy.data = np.round(sparse_row_scipy.data * 100).astype(np.uint32)
    return sparse_row_scipy


def _get_numba_matrix_attr(sparse_row_scipy):
    values = sparse_row_scipy.data.astype(np.uint32)
    col_index = sparse_row_scipy.indices.astype(np.uint32)
    row_index = sparse_row_scipy.indptr.astype(np.uint32)
    return values, col_index, row_index



import os
import sys
import threading
from typing import Iterable, Literal, Optional

from loguru import logger

LogLevel = Literal["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class Config:
    """
    Global settings: Overpass endpoint(s), timeouts, logging, tag sets, and rate limiting.
    All setters validate inputs; logging can be configured to stderr and/or file.
    """

    def __init__(self):
        # --- Overpass endpoints ---
        self.overpass_url: str = os.getenv("OVERPASS_URL", "https://overpass-api.de/api/interpreter")
        self.user_agent: str = os.getenv("OVERPASS_USER_AGENT", "iduedu/0.1 (+https://example.org)")
        self.proxies: Optional[dict[str, str]] = None
        self.verify_ssl: bool = True

        # --- Networking / retries ---
        self.timeout: int = int(os.getenv("OVERPASS_TIMEOUT", "120"))
        self.overpass_min_interval: float = float(os.getenv("OVERPASS_MIN_INTERVAL", "2"))
        self.overpass_max_retries: int = int(os.getenv("OVERPASS_MAX_RETRIES", "3"))
        self.overpass_retry_statuses: tuple[int, ...] = (429, 502, 503, 504)
        self.overpass_backoff_base: float = float(os.getenv("OVERPASS_BACKOFF_BASE", "0.5"))

        # --- UX ---
        self.enable_tqdm_bar: bool = os.getenv("ENABLE_TQDM", "1") not in {"0", "false", "False"}

        # --- Tags to keep on edges ---
        self.drive_useful_edges_attr = frozenset({"highway", "name", "lanes"})
        self.walk_useful_edges_attr = frozenset({"highway", "name"})
        self.transport_useful_edges_attr = frozenset({"name"})

        # --- Logging ---
        self.logger = logger
        self.configure_logging(level=os.getenv("LOG_LEVEL", "INFO").upper())

        # --- Internals ---
        self._lock = threading.RLock()

    # ------------- Logging -------------
    def configure_logging(self, level: LogLevel = "INFO", *, to_file: str | None = None, json: bool = False):
        """
        Configure Loguru: stderr sink (+optional file sink), level, formatting.
        """
        self.logger.remove()
        fmt = "{time:YYYY-MM-DD HH:mm:ss.SSS} | {level:<8} | {message}" if not json else "{message}"
        self.logger.add(sys.stderr, level=level, backtrace=True, diagnose=False, format=fmt)
        if to_file:
            self.logger.add(
                to_file,
                level=level,
                rotation="20 MB",
                retention="14 days",
                compression="zip",
                enqueue=True,
                backtrace=True,
                diagnose=False,
                format=fmt,
            )

    # ------------- Overpass control -------------
    def set_overpass_url(self, url: str):
        """
        Set active Overpass URL (validated), keeps the pool intact.
        """
        if not (url.startswith("http://") or url.startswith("https://")):
            raise ValueError(f"Invalid Overpass URL: {url!r}")
        with self._lock:
            self.overpass_url = url

    def set_timeout(self, timeout: int):
        if timeout <= 0:
            raise ValueError("timeout must be > 0")
        self.timeout = int(timeout)

    def set_enable_tqdm(self, enable: bool):
        self.enable_tqdm_bar = bool(enable)

    def set_drive_useful_edges_attr(self, attr: Iterable[str]):
        self.drive_useful_edges_attr = frozenset(attr)

    def set_walk_useful_edges_attr(self, attr: Iterable[str]):
        self.walk_useful_edges_attr = frozenset(attr)

    def set_transport_useful_edges_attr(self, attr: Iterable[str]):
        self.transport_useful_edges_attr = frozenset(attr)

    def set_rate_limit(
        self, *, min_interval: float | None = None, max_retries: int | None = None, backoff_base: float | None = None
    ):
        if min_interval is not None:
            if min_interval < 0:
                raise ValueError("min_interval must be >= 0")
            self.overpass_min_interval = float(min_interval)
        if max_retries is not None:
            if max_retries < 0:
                raise ValueError("max_retries must be >= 0")
            self.overpass_max_retries = int(max_retries)
        if backoff_base is not None:
            if backoff_base <= 0:
                raise ValueError("backoff_base must be > 0")
            self.overpass_backoff_base = float(backoff_base)

    def to_dict(self) -> dict:
        return {
            "overpass_url": self.overpass_url,
            "timeout": self.timeout,
            "enable_tqdm_bar": self.enable_tqdm_bar,
            "drive_useful_edges_attr": sorted(self.drive_useful_edges_attr),
            "walk_useful_edges_attr": sorted(self.walk_useful_edges_attr),
            "transport_useful_edges_attr": sorted(self.transport_useful_edges_attr),
            "overpass_min_interval": self.overpass_min_interval,
            "overpass_max_retries": self.overpass_max_retries,
            "overpass_retry_statuses": self.overpass_retry_statuses,
            "overpass_backoff_base": self.overpass_backoff_base,
            "user_agent": self.user_agent,
            "proxies": self.proxies,
            "verify_ssl": self.verify_ssl,
        }

    @classmethod
    def from_env(cls) -> "Config":
        """
        Construct Config using environment variables where available.
        """
        cfg = cls()
        return cfg


config = Config()



import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from shapely import LineString, MultiPolygon, Polygon
from tqdm.auto import tqdm
from tqdm.contrib.concurrent import process_map, thread_map

from iduedu import config
from iduedu.enums.pt_enums import PublicTrasport
from iduedu.modules.overpass_parsers import parse_overpass_subway_data, parse_overpass_to_edgenode

from .graph_transformers import clip_nx_graph, estimate_crs_for_bounds
from .overpass_downloaders import (
    get_4326_boundary,
    get_routes_by_poly,
)

logger = config.logger


def _graph_data_to_nx(graph_df, keep_geometry: bool = True, additional_data=None) -> nx.DiGraph:
    """
    Build a directed public-transport graph from a mixed edge/node DataFrame, with optional subway add-ons.

    Input `graph_df` contains both node rows (non-null `node_id`) and edge rows (null `node_id`).
    Nodes are grouped/merged (platforms by rounded coords; other types by (coord, route, type, then by ref_id)),
    then edges are mapped to these merged nodes. If `additional_data` (subway entrances/transfers) is provided,
    edges with `u_ref`/`v_ref` are joined by `ref_id`.

    Parameters:
        graph_df (pd.DataFrame): Mixed table with:
            node rows → columns: `node_id`, `point` (projected meters), `route`, `type`, `ref_id`, `extra_data`;
            edge rows → columns: `u`, `v`, `type`, `extra_data`, `route`, `geometry`.
        keep_geometry (bool): If True, store shapely `geometry` on edges; otherwise drop it.
        additional_data (tuple | None): Optional `(additional_edges, additional_nodes)` produced by
            `parse_overpass_subway_data`. `additional_edges` must have `u_ref`, `v_ref`, `type`, optional
            `geometry`, `route`, `extra_data`; `additional_nodes` must have `ref_id`, `point`, `type`, `extra_data`.

    Returns:
        (nx.DiGraph): Directed graph
    """

    def avg_point(points):
        arr = np.asarray([p for p in points if p is not None], dtype=float)
        return tuple(arr.mean(axis=0)) if len(arr) else None

    def merge_dicts_last(dicts):
        out = {}
        for d in dicts.dropna():
            for k, v in d.items():
                out[k] = v
        return out

    nodes_col = ["node_id", "point", "route", "type", "ref_id", "extra_data"]
    edges_col = ["u", "v", "type", "extra_data", "route", "geometry"]
    graph_nodes = graph_df[~graph_df["node_id"].isna()][nodes_col].copy()
    graph_edges = graph_df[graph_df["node_id"].isna()][edges_col].copy()

    if additional_data is not None:
        additional_edges, additional_nodes = additional_data
        graph_nodes_combined = graph_nodes.merge(additional_nodes, left_on="ref_id", right_on="ref_id", how="outer")
        for column in nodes_col:
            if column in graph_nodes_combined.columns:
                continue
            else:
                graph_nodes_combined[column] = graph_nodes_combined[f"{column}_y"].combine_first(
                    graph_nodes_combined[f"{column}_x"]
                )
        graph_nodes_combined = graph_nodes_combined[nodes_col]

        no_point_refs = graph_nodes_combined[
            (graph_nodes_combined["point"].isna()) & (~graph_nodes_combined["ref_id"].isna())
        ][["ref_id"]].copy()
        if len(no_point_refs) > 0:
            no_point_refs = no_point_refs.merge(
                additional_edges[["v_ref", "u_ref"]], left_on="ref_id", right_on="v_ref"
            )[["v_ref", "u_ref"]].drop_duplicates(subset="u_ref")
            no_point_refs = no_point_refs.merge(
                graph_nodes[["node_id", "type", "ref_id"]], left_on="u_ref", right_on="ref_id"
            ).drop(columns=["ref_id"])
            no_point_refs = no_point_refs.merge(
                graph_edges[["u", "v"]], left_on="node_id", right_on="u", how="left"
            ).drop(columns=["u"])
            no_point_refs = no_point_refs.merge(
                graph_nodes[["node_id", "type", "ref_id"]].add_prefix("potential_"),
                left_on="v",
                right_on="potential_node_id",
            )
            no_point_refs = no_point_refs[no_point_refs["potential_type"] == "platform"]
            remap_refs = no_point_refs.set_index("potential_ref_id")["v_ref"].to_dict()
            s = graph_nodes_combined["ref_id"].copy()
            mapped = s.map(remap_refs)
            mask = mapped.notna()
            graph_nodes_combined.loc[mask, "ref_id"] = mapped[mask]
            graph_nodes_combined.loc[mask, "type"] = "subway_platform"

            graph_nodes = graph_nodes_combined.dropna(subset=["node_id", "point"], how="all").copy()
            graph_nodes["route"] = graph_nodes["route"].fillna("subway_transit")
            # TODO delete duplicated platform node

    platforms = graph_nodes[graph_nodes["type"] == "platform"].copy()
    platforms["point_group"] = platforms["point"].apply(lambda p: (round(p[0]), round(p[1])))
    platforms = platforms.groupby("point_group", as_index=False).agg(
        point=("point", "first"),
        node_id=("node_id", lambda s: tuple(s.dropna())),
        route=("route", lambda s: tuple(s)),
        ref_id=("ref_id", lambda s: tuple(s.dropna())),
        extra_data=("extra_data", merge_dicts_last),
    )
    platforms["type"] = "platform"

    not_platforms = graph_nodes[graph_nodes["type"] != "platform"].copy()
    not_platforms["point_group"] = not_platforms["point"].apply(lambda p: (round(p[0]), round(p[1])))
    not_platforms = not_platforms.groupby(["point_group", "route", "type"], as_index=False).agg(
        point=("point", "first"),
        node_id=("node_id", lambda s: tuple(s.dropna())),
        ref_id=("ref_id", lambda s: tuple(set(s.dropna()))),
        extra_data=("extra_data", merge_dicts_last),
    )
    not_platforms = not_platforms.groupby(["ref_id", "route", "type"], as_index=False).agg(
        point=("point", avg_point),
        node_id=("node_id", "sum"),
        extra_data=("extra_data", merge_dicts_last),
    )

    all_nodes = pd.concat([platforms, not_platforms], ignore_index=True)

    map_nodeid_to_idx = {}
    map_refid_to_idx = {}

    for idx, row in all_nodes.iterrows():
        for nid in row["node_id"] or []:
            map_nodeid_to_idx[nid] = idx
        rids = row["ref_id"]
        if not isinstance(rids, (list, tuple)):
            rids = [rids]
        for rid in rids or []:
            map_refid_to_idx[rid] = idx

    edges_existing = graph_edges[["route", "type", "u", "v", "geometry", "extra_data"]].copy()
    edges_existing["u"] = edges_existing["u"].map(map_nodeid_to_idx)
    edges_existing["v"] = edges_existing["v"].map(map_nodeid_to_idx)
    edges_existing = edges_existing.dropna(subset=["u", "v"]).copy()

    if additional_data is not None:
        additional_edges["u"] = additional_edges["u_ref"].map(map_refid_to_idx)
        additional_edges["v"] = additional_edges["v_ref"].map(map_refid_to_idx)
        additional_edges = additional_edges.dropna(subset=["u", "v"]).copy()

        def _ensure_geom(row):
            if row.get("geometry") is not None:
                return row["geometry"]
            pu = all_nodes.loc[int(row["u"]), "point"]
            pv = all_nodes.loc[int(row["v"]), "point"]
            return LineString([pu, pv])

        additional_edges["geometry"] = additional_edges.apply(_ensure_geom, axis=1)
    else:
        additional_edges = pd.DataFrame()

    edges = pd.concat([edges_existing, additional_edges], ignore_index=True)

    if "length_meter" not in edges.columns:
        edges["length_meter"] = np.nan
    if "time_min" not in edges.columns:
        edges["time_min"] = np.nan

    def calc_len_time(row):
        if row.type == "boarding":
            return 0.0, 0.0
        geom = row.geometry
        if geom is None:
            return 0.0, 0.0
        length = float(round(geom.length, 3))
        speed = PublicTrasport[row.type.upper()].avg_speed
        return length, float(round(length / speed, 3))

    mask_missing = edges["length_meter"].isna() | edges["time_min"].isna()

    vals = edges.loc[mask_missing].apply(calc_len_time, axis=1, result_type="expand")
    vals.columns = ["length_meter", "time_min"]
    edges.loc[mask_missing, ["length_meter", "time_min"]] = vals

    graph = nx.DiGraph()

    for idx, node in all_nodes.iterrows():
        route = list(set(node["route"])) if isinstance(node["route"], tuple) else [node["route"]]
        if len(route) == 1:
            route = route[0]
        graph.add_node(
            idx,
            x=float(node["point"][0]),
            y=float(node["point"][1]),
            type=node["type"],
            route=route,
            ref_id=(node["ref_id"][0] if isinstance(node["ref_id"], tuple) and node["ref_id"] else node["ref_id"]),
            **(node["extra_data"] if isinstance(node["extra_data"], dict) else {}),
        )

    for _, e in edges.iterrows():
        graph.add_edge(
            int(e["u"]),
            int(e["v"]),
            route=e["route"],
            type=e["type"],
            geometry=(e["geometry"] if keep_geometry else None),
            length_meter=e["length_meter"],
            time_min=e["time_min"],
            **(e["extra_data"] if isinstance(e["extra_data"], dict) else {}),
        )

    return graph


def _multi_get_routes_by_poly(args):
    return get_routes_by_poly(*args)


def _multi_parse_overpass_to_edgenode(args):
    return parse_overpass_to_edgenode(*args)


def _get_public_transport_graph(
    osm_id: int,
    territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None,
    transport_types: list[str],
    osm_edge_tags: list[str],
    clip_by_territory: bool = False,
    keep_edge_geometry: bool = True,
):
    """
    Build a directed public-transport graph for one or several transport types inside a territory.

    The function:
    1) resolves a boundary polygon (EPSG:4326);
    2) downloads routes via Overpass in parallel;
    3) (for subway) additionally parses stop areas/groups and stations, producing entrances/transfers;
    4) parses each route into edge/node rows (parallel for large inputs);
    5) assembles a single `nx.DiGraph` via `_graph_data_to_nx`, computing missing edge length/time;
    6) optionally clips the graph by the territory.

    Parameters:
        osm_id (int): OSM relation/area id of the territory; used if `territory` is not provided.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Boundary geometry in EPSG:4326.
        transport_types (list[str]): Transport types to include (e.g., `["bus", "tram", "subway"]`).
        osm_edge_tags (list[str]): Which route/member tags to retain on edges/nodes (overrides defaults).
        clip_by_territory (bool): If True, clip the final graph to the (projected) boundary.
        keep_edge_geometry (bool): If True, store shapely `geometry` on edges.

    Returns:
        (nx.DiGraph): Directed PT graph. Graph attributes set by this function:
            - `graph["crs"]` (int/EPSG of the local projected CRS),
            - `graph["type"]` = "public_trasport" (sic).
    """

    polygon = get_4326_boundary(osm_id=osm_id, territory=territory)

    args_list = [(polygon, transport) for transport in transport_types]

    # Если парсим метро - ожидаем в ответе информацию о станциях
    platform_stop_data_use = False
    if "subway" in transport_types:
        platform_stop_data_use = True

    if not config.enable_tqdm_bar:
        logger.debug("Downloading pt routes")
    overpass_data = pd.concat(
        thread_map(
            _multi_get_routes_by_poly,
            args_list,
            desc="Downloading public transport routes from OSM",
            disable=not config.enable_tqdm_bar,
        ),
        ignore_index=True,
    ).reset_index(drop=True)

    if overpass_data.shape[0] == 0:
        logger.warning("No routes found for public transport.")
        return nx.Graph()

    local_crs = estimate_crs_for_bounds(*polygon.bounds).to_epsg()

    # необходимые osm теги из relation маршрута
    if osm_edge_tags is None:
        needed_tags = set(config.transport_useful_edges_attr)
    else:
        needed_tags = set(osm_edge_tags)

    # Отделяем станции от маршрутов при необходимости
    if platform_stop_data_use:
        for column in ["is_stop_area", "is_stop_area_group", "is_station"]:
            overpass_data[column] = overpass_data[column].astype("boolean").fillna(False)
        routes_data = overpass_data[
            ~((overpass_data["is_stop_area"]) | (overpass_data["is_stop_area_group"]) | (overpass_data["is_station"]))
        ].copy()
        stop_areas = overpass_data[overpass_data["is_stop_area"]].copy()
        stop_areas_group = overpass_data[overpass_data["is_stop_area_group"]].copy()
        stations_data = overpass_data[overpass_data["is_station"]].copy()
        add_data = parse_overpass_subway_data(stop_areas, stop_areas_group, stations_data, local_crs)
    else:
        routes_data = overpass_data.copy()
        add_data = None

    if not config.enable_tqdm_bar:
        logger.debug("Parsing public transport routes")
    if overpass_data.shape[0] > 100:
        # Если много маршрутов - обрабатываем в параллели
        edgenode_for_routes = process_map(
            _multi_parse_overpass_to_edgenode,
            [(row, local_crs, needed_tags) for _, row in routes_data.iterrows()],
            desc="Parsing public transport routes",
            chunksize=1,
            disable=not config.enable_tqdm_bar,
        )
    else:
        tqdm.pandas(desc="Parsing public transport routes", disable=not config.enable_tqdm_bar)
        edgenode_for_routes = [
            data
            for data in routes_data.progress_apply(
                lambda x: parse_overpass_to_edgenode(x, local_crs, needed_tags), axis=1
            ).tolist()
            if data is not None
        ]

    if len(edgenode_for_routes) == 0:
        logger.warning("No routes were parsed for public transport.")
        return nx.DiGraph()
    graph_df = pd.concat(edgenode_for_routes, ignore_index=True)
    to_return = _graph_data_to_nx(graph_df, keep_geometry=keep_edge_geometry, additional_data=add_data)
    to_return.graph["crs"] = local_crs
    to_return.graph["type"] = "public_trasport"

    if clip_by_territory:
        polygon = gpd.GeoSeries([polygon], crs=4326).to_crs(local_crs).union_all()
        to_return = clip_nx_graph(to_return, polygon)

    logger.debug("Done!")
    return to_return


def get_single_public_transport_graph(
    public_transport_type: str | PublicTrasport,
    *,
    osm_id: int | None = None,
    territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None = None,
    clip_by_territory: bool = False,
    keep_edge_geometry: bool = True,
    osm_edge_tags: list[str] | None = None,  # overrides default tags
):
    """
    Build a directed graph for a single public-transport mode within a given territory.

    The function resolves a boundary (by `osm_id` or `territory`), downloads OpenStreetMap routes  inside that boundary,
    converts them into a projected `nx.DiGraph`, and computes per-edge length (meters) and travel time (minutes).
    When the mode is **subway**, additional station context is incorporated: entrances/exits and inter-station
    transfers are added; station metadata (e.g., name, depth) is attached to nodes where available.

    Parameters:
        public_transport_type (str | PublicTrasport): One mode, e.g. `"bus"`, `"tram"`, `"trolleybus"`, `"subway"`.
            You may pass a `PublicTrasport` enum or its string value.
        osm_id (int | None): OSM relation/area id of the territory. Provide this or `territory`.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Boundary geometry in EPSG:4326 (or a GeoDataFrame).
            Used when `osm_id` is not given.
        clip_by_territory (bool): If True, the resulting graph is clipped to the boundary (in the local CRS).
        keep_edge_geometry (bool): If True, edge shapes (`shapely` geometries in local CRS) are stored on edges.
        osm_edge_tags (list[str] | None): Subset of OSM tags to retain on edges/nodes. If None, a sensible default
            is used from configuration; only requested keys are joined from OSM element tags.

    Returns:
        (nx.DiGraph): Directed PT graph with:
            - node attrs: `x`, `y` (floats, local CRS), `type`, `route`, `ref_id`, plus merged station `extra_data` (if any);
            - edge attrs: `type`, `route`, `length_meter`, `time_min`, optional `geometry`, and selected OSM tags.

          Graph attrs: `graph["crs"]` (EPSG int of the local projected CRS), `graph["type"]` = `"public_trasport"`.

    Notes:
        - Lengths and times are computed in a **local projected CRS** estimated from the boundary; per-edge speeds are
          taken from mode-specific defaults (and, for subway connectors, from connector-type defaults).
    """
    public_transport_type = (
        public_transport_type.value() if isinstance(public_transport_type, PublicTrasport) else public_transport_type
    )
    return _get_public_transport_graph(
        osm_id=osm_id,
        territory=territory,
        transport_types=[public_transport_type],
        osm_edge_tags=osm_edge_tags,
        clip_by_territory=clip_by_territory,
        keep_edge_geometry=keep_edge_geometry,
    )


def get_all_public_transport_graph(
    *,
    osm_id: int | None = None,
    territory: Polygon | MultiPolygon | gpd.GeoDataFrame | None = None,
    clip_by_territory: bool = False,
    keep_edge_geometry: bool = True,
    transport_types: list[PublicTrasport] = None,
    osm_edge_tags: list[str] | None = None,  # overrides default tags
) -> nx.Graph:
    """
    Build a combined directed graph for multiple public-transport modes within a territory.

    The function collects routes for the requested modes (by default: **tram**, **bus**, **trolleybus**, **subway**),
    converts them into a single projected graph, and computes per-edge length and time. Edges from different modes
    coexist in the same `nx.DiGraph`; node ids are unified across modes. For the **subway** mode, station context
    is added (entrances/exits and inter-station transfers), and available station metadata is merged into node attrs.

    Parameters:
        osm_id (int | None): OSM relation/area id of the territory. Provide this or `territory`.
        territory (Polygon | MultiPolygon | gpd.GeoDataFrame | None): Boundary geometry in EPSG:4326 (or a GeoDataFrame).
        clip_by_territory (bool): If True, clip the final graph to the boundary (in the local CRS).
        keep_edge_geometry (bool): If True, retain `shapely` geometries on edges.
        transport_types (list[PublicTrasport] | None): List of modes to include. Defaults to
            `[PublicTrasport.TRAM, PublicTrasport.BUS, PublicTrasport.TROLLEYBUS, PublicTrasport.SUBWAY]`.
            All items must be `PublicTrasport` enums.
        osm_edge_tags (list[str] | None): Which OSM tags to keep on edges/nodes. If None, a default subset from
            configuration is used; only these keys are joined from OSM.

    Returns:
        (nx.DiGraph): Combined directed PT graph. Typical attributes:
            - node attrs: `x`, `y` (local CRS), `type`, `route`, `ref_id`, station `extra_data` where applicable;
            - edge attrs: `type`, `route`, `length_meter`, `time_min`, optional `geometry`, plus selected OSM tags.

          Graph attrs: `graph["crs"]` (EPSG int), `graph["type"]` = `"public_trasport"`.

    Notes:
        Each mode’s ways are downloaded inside the boundary and transformed into directed edges; per-edge speeds are
        taken from mode-specific defaults (and, for subway connectors, from connector-type defaults).
    """

    if transport_types is None:
        transport_types = [PublicTrasport.TRAM, PublicTrasport.BUS, PublicTrasport.TROLLEYBUS, PublicTrasport.SUBWAY]
    else:
        for transport_type in transport_types:
            if not isinstance(transport_type, PublicTrasport):
                raise ValueError(f"transport_type {transport_type} is not a valid transport type.")

    transports = [transport.value for transport in transport_types if isinstance(transport, PublicTrasport)]

    return _get_public_transport_graph(
        osm_id=osm_id,
        territory=territory,
        transport_types=transports,
        osm_edge_tags=osm_edge_tags,
        clip_by_territory=clip_by_territory,
        keep_edge_geometry=keep_edge_geometry,
    )



from collections import OrderedDict

import numba as nb
import numpy as np
from numba.core.types import float32, uint8, uint32

numba_csr_val_types = [float32[:], uint32[:], uint8[:]]


def __build_csr_cls(nb_type):

    spec_csr_matrix = [
        ("values", nb_type),
        ("col_index", nb.types.uint32[:]),
        ("row_index", nb.types.uint32[:]),
        ("nnz_rows", nb.types.uint32[:]),
        ("tot_rows", uint32),
    ]
    spec_csr_matrix = OrderedDict(spec_csr_matrix)

    @nb.experimental.jitclass(spec_csr_matrix)
    class CSRMatrix:  # pragma: no cover
        """
        a minimal CSR matrix implementation
        get_nnz and get_row should only be used on rows for which a value is present
        otherwise IndexErrors will be raised

        """

        def __init__(self, values, col_index, row_index):
            self.values = values
            self.col_index = col_index
            self.row_index = row_index
            self.nnz_rows = self.__set_nnz_rows()
            self.tot_rows = len(row_index) - 1

        def get_nnz(self, row):
            row_start = self.row_index[row]
            row_end = self.row_index[row + 1]
            return self.col_index[row_start:row_end]

        def get_row(self, row):
            row_start = self.row_index[row]
            row_end = self.row_index[row + 1]
            return self.values[row_start:row_end]

        def __set_nnz_rows(self):
            rows = []
            for row in np.arange(len(self.row_index[:-1]), dtype=np.uint32):
                if len(self.get_nnz(row)) > 0:
                    rows.append(row)
            return np.array(rows, dtype=np.uint32)

        def get_nnz_rows(self):
            # get rows that have non-zero values
            return self.nnz_rows

    return CSRMatrix


UI32CSRMatrix = __build_csr_cls(nb.uint32[:])

try:
    ui32csr_type = UI32CSRMatrix.class_type.instance_type
except Exception:
    ui32csr_type = None
